# run test python -m unittest discover tests

from eticas.metrics.base_metric import BaseFairnessMetric
import pandas as pd
import unittest


class TestFairnessMetric(BaseFairnessMetric):

    def compute(self, input_data):
        return super().compute(input_data)


class TestBaseFairnessMetric(unittest.TestCase):

    def setUp(self):
        # Create an instance of the dummy metric
        self.metric = TestFairnessMetric()
        # Create a simple DataFrame for testing

    def test_compute(self):

        # Attempting to call compute should raise NotImplementedError.
        with self.assertRaises(NotImplementedError) as context:
            self.metric.compute(pd.DataFrame())
        # Optional: Assert that the exception message is as expected.
        self.assertEqual(str(context.exception), "Not implemented.")

    def test_normalize_value(self):
        # Test normalization function.
        with self.assertRaises(NotImplementedError) as context:
            self.metric.compute(self.metric.normalize_value(50))
        # Optional: Assert that the exception message is as expected.
        self.assertEqual(str(context.exception), "Not implemented.")

    def test_get_bias_level(self):
        # Test bias level determination based on normalized value.
        self.assertEqual(self.metric.get_bias_level(85),
                         'Low/no bias', "Value above 80 should result in 'Low/no bias'.")
        self.assertEqual(self.metric.get_bias_level(75),
                         'Medium bias', "Value between 61 and 80 should result in 'Medium bias'.")
        self.assertEqual(self.metric.get_bias_level(60),
                         'High bias', "Value of 60 or less should result in 'High bias'.")
        self.assertEqual(self.metric.get_bias_level(30),
                         'High bias', "Low value should result in 'High bias'.")

    def test_validate_parameters_success(self):
        # Should not raise an error when valid parameters are provided.
        try:
            self.metric.validate_parameters(input_data=1, sensitive_attrs="test", input_features=['age'])
        except ValueError:
            self.fail("validate_parameters raised ValueError unexpectedly with valid parameters.")

    def test_validate_parameters_failure(self):
        # Should raise a ValueError when a parameter is None.
        with self.assertRaises(ValueError) as context:
            self.metric.validate_parameters(input_data=None)
        self.assertIn("You must provide a value for input_data", str(context.exception))

    def test_str_method(self):
        # Test that the __str__ method returns the class name.
        self.assertEqual(str(self.metric), "TestFairnessMetric", "The __str__ method should return the class name.")


if __name__ == '__main__':
    unittest.main()



import os
import requests
import pandas as pd
import numpy as np
import json


def get_audit(audit_id=2289):
    """
    Parameters
    ----------
   :param audit_id: The ID of the audit to be explored.
    Returns
    ----------
   :return: A dictionary containing the results for each group
    """
    BASE_URL = os.getenv("ITACA_BASE_URL")
    if not BASE_URL:
        raise ValueError("❌ 'ITACA_BASE_URL' NO DEFINED.")
    API_TOKEN = os.getenv("ITACA_API_TOKEN")
    if not API_TOKEN:
        raise ValueError("❌ 'ITACA_API_TOKEN' NO DEFINED.")

    url = f"{BASE_URL}audit/{audit_id}/"
    headers = {
        "Authorization": f"Token {API_TOKEN}",
        "Accept": "application/json"
    }
    response = requests.get(url, headers=headers)
    if response.status_code == 200:
        return response.json()
    else:
        raise requests.HTTPError(f"❌ Error {response.status_code}: {response.text}")


def get_departments():
    """
    Returns
    ----------
   :return: A DataFrame with all the departments.
    """
    BASE_URL = os.getenv("ITACA_BASE_URL")
    if not BASE_URL:
        raise ValueError("❌ 'ITACA_BASE_URL' NO DEFINED.")
    API_TOKEN = os.getenv("ITACA_API_TOKEN")
    if not API_TOKEN:
        raise ValueError("❌ 'ITACA_API_TOKEN' NO DEFINED.")
    url = f"{BASE_URL}department/"
    headers = {
        "Authorization": f"Token {API_TOKEN}",
        "Accept": "application/json"
    }
    response = requests.get(url, headers=headers)
    if response.status_code == 200:
        return pd.DataFrame(response.json())
    else:
        raise requests.HTTPError(f"❌ Error {response.status_code}: {response.text}")


def get_models(department=None):
    """
    Parameters
    ----------
   :param department: The ID of the department to be explored.
    Returns
    ----------
   :return: A DataFrame containing all the models for selected department.
    """
    BASE_URL = os.getenv("ITACA_BASE_URL")
    if not BASE_URL:
        raise ValueError("❌ 'ITACA_BASE_URL' NO DEFINED.")
    API_TOKEN = os.getenv("ITACA_API_TOKEN")
    if not API_TOKEN:
        raise ValueError("❌ 'ITACA_API_TOKEN' NO DEFINED.")
    url = f"{BASE_URL}department/{department}/model/"
    headers = {
        "Authorization": f"Token {API_TOKEN}",
        "Accept": "application/json"
    }
    response = requests.get(url, headers=headers)
    if response.status_code == 200:
        return pd.DataFrame(response.json())
    else:
        raise requests.HTTPError(f"❌ Error {response.status_code}: {response.text}")


def get_audits(model=None):
    """
    Parameters
    ----------
   :param model: The ID of the model to be explored.
    Returns
    ----------
   :return: A DataFrame containing the audits.
    """
    BASE_URL = os.getenv("ITACA_BASE_URL")
    if not BASE_URL:
        raise ValueError("❌ 'ITACA_BASE_URL' NO DEFINED.")
    API_TOKEN = os.getenv("ITACA_API_TOKEN")
    if not API_TOKEN:
        raise ValueError("❌ 'ITACA_API_TOKEN' NO DEFINED.")
    url = f"{BASE_URL}audit/?auditable_model_id={model}"
    headers = {
        "Authorization": f"Token {API_TOKEN}",
        "Accept": "application/json"
    }
    response = requests.get(url, headers=headers)
    if response.status_code == 200:
        result = response.json()['results']
        return pd.DataFrame(result)
    else:
        raise requests.HTTPError(f"❌ Error {response.status_code}: {response.text}")


def upload_audit(department_id=None,
                 model_id=None,
                 model=None):
    BASE_URL = os.getenv("ITACA_BASE_URL")
    if not BASE_URL:
        raise ValueError("❌ 'ITACA_BASE_URL' NO DEFINED.")
    API_TOKEN = os.getenv("ITACA_API_TOKEN")
    if not API_TOKEN:
        raise ValueError("❌ 'ITACA_API_TOKEN' NO DEFINED.")
    departments = get_departments()
    if departments.shape[0] > 0:
        if department_id not in departments.id.values.tolist():
            raise ValueError("Deparment ID does not exist..")
    else:
        raise ValueError("Deparment ID does not exist..")
    models = get_models(department=department_id)
    if models.shape[0] > 0:
        if model_id not in models.id.values.tolist():
            raise ValueError("Model ID does not exist..")
    else:
        raise ValueError("Model ID does not exist..")
    url = f"{BASE_URL}department/{department_id}/model/{model_id}/audit/result"
    audit_metrics = upload_json_audit(model)
    json_str = json.dumps(audit_metrics)
    files = {
        "metrics": ("archivo.json", json_str, "multipart/form-data")
    }
    headers = {
        "Authorization": f"Token {API_TOKEN}"
    }
    response = requests.post(url, headers=headers, files=files)

    return response.status_code


def overview(key0, key1,
             no_none_share,
             ref_evol):
    result = {}
    for s in no_none_share:
        diff = np.round(s[0]-ref_evol, 2).item()
        result.update({key0+'_'+s[1]+'_difference': abs(diff)})
        result.update({key0+'_'+s[1]+'_evol': 'equal to' if diff == 0 else 'below' if diff < 0 else 'above'})

    if len(no_none_share) > 1:
        for n in range(len(no_none_share)-1):
            diff = no_none_share[n+1][0] - no_none_share[n][0]
            key = f"{key1}_{no_none_share[n][1]}_{no_none_share[n+1][1]}"
            value = ['negative', 'neutral', 'positive'][(diff > 0) - (diff < 0) + 1]
            result.update({key: value})
    return result


def scoring_evolution(first_share,
                      last_share,
                      ref_share):
    if last_share >= ref_share:
        share_risk = 100
    elif last_share < first_share:
        penalty = -20
        norm_share = (last_share - 0) / (100 - 0)
        share_risk = max(np.round(penalty + (norm_share * 100), 4).item(), 0)
    else:
        norm_share = (last_share - 0) / (100 - 0)
        share_risk = max(np.round((norm_share * 100), 4).item(), 0)

    return share_risk


def normalize_benchmarking(X):
    result = X - ((X - 1) * 2) if X > 1 else X
    result = max(result, 0)
    result = result * 100
    return result


def bias_direction(X):
    if X == 1:
        return 'Correct-representation'
    if X < 1:
        return 'Under-representation'
    else:
        return 'Over-representation'


def upload_json_audit(model):
    json_result = model.json_results(norm_values=True)
    audit_result = {}
    for p in json_result.keys():
        if p != 'error':
            if (p.lower() not in ['gender', 'ethnicity', 'age', 'gender_ethnicity']) & ('sensitive_' not in p.lower()):
                p_id = 'sensitive_'+p.lower()
            else:
                p_id = ''+p.lower()
            if p in str(model.distribution_ref):
                ref_aux = model.distribution_ref[p]
            else:
                ref_aux = 50
            audit_result[p_id] = {}
            training_data = json_result.get(p, {}).get('benchmarking',
                                                       {}).get('labeled_da_inconsistency', None)
            training_positive = json_result.get(p, {}).get('benchmarking',
                                                           {}).get('labeled_da_positive', None)
            operational_data = json_result.get(p, {}).get('benchmarking',
                                                          {}).get('operational_da_inconsistency', None)
            operational_positive = json_result.get(p, {}).get('benchmarking',
                                                              {}).get('operational_da_positive', None)
            impact_data = json_result.get(p, {}).get('benchmarking',
                                                     {}).get('impact_da_inconsistency', None)
            impact_positive = json_result.get(p, {}).get('benchmarking',
                                                         {}).get('impact_da_positive', None)

            audit_result[p_id]['benchmarking'] = {
                'ref': ref_aux,
                'training_data': training_data,
                'training_positive': training_positive,
                'operational_data': operational_data,
                'operational_positive':  operational_positive,
                'impact_data': impact_data,
                'impact_positive':  impact_positive,

                'training_data_risk': normalize_benchmarking(
                    0 if training_data is None else training_data/ref_aux),
                'training_positive_risk':  normalize_benchmarking(
                    0 if training_positive is None else training_positive/ref_aux),
                'operational_data_risk': normalize_benchmarking(
                    0 if operational_data is None else operational_data/ref_aux),
                'operational_positive_risk':  normalize_benchmarking(
                    0 if operational_positive is None else operational_positive/ref_aux),
                'impact_data_risk': normalize_benchmarking(
                    0 if impact_data is None else impact_data/ref_aux),
                'impact_positive_risk': normalize_benchmarking(
                    0 if impact_positive is None else impact_positive/ref_aux),

                'training_data_direction': bias_direction(
                    0 if training_data is None else training_data/ref_aux),
                'training_positive_direction':  bias_direction(
                    0 if training_positive is None else training_positive/ref_aux),
                'operational_data_direction': bias_direction(
                    0 if operational_data is None else operational_data/ref_aux),
                'operational_positive_direction':  bias_direction(
                    0 if operational_positive is None else operational_positive/ref_aux),
                'impact_data_direction': bias_direction(
                    0 if impact_data is None else impact_data/ref_aux),
                'impact_positive_direction': bias_direction(
                    0 if impact_positive is None else impact_positive/ref_aux),
                }

            audit_result[p_id]['distribution'] = {
                'ref': 80,
                'training_proxy': json_result.get(p,
                                                  {}).get('distribution',
                                                          {}).get('labeled_dxa_inconsistency', None),
                'training_label': json_result.get(p,
                                                  {}).get('distribution',
                                                          {}).get('labeled_da_informative', None),
                'operational_proxy': json_result.get(p,
                                                     {}).get('distribution',
                                                             {}).get('operational_dxa_inconsistency', None),
                'operational_label': json_result.get(p,
                                                     {}).get('distribution',
                                                             {}).get('operational_da_informative', None),
                'impact_proxy': json_result.get(p,
                                                {}).get('distribution',
                                                        {}).get('impact_dxa_inconsistency', None),
                'impact_label': json_result.get(p,
                                                {}).get('distribution',
                                                        {}).get('impact_da_informative', None),
                }
            audit_result[p_id]['fairness'] = {
                'ref': 80,

                'training_DI': json_result.get(p,
                                               {}).get('fairness',
                                                       {}).get('labeled_DI', None),
                'operational_DI': json_result.get(p,
                                                  {}).get('fairness',
                                                          {}).get('operational_DI', None),
                'impact_DI': json_result.get(p,
                                             {}).get('fairness',
                                                     {}).get('impact_DI', None),

                'training_SPD': json_result.get(p,
                                                {}).get('fairness',
                                                        {}).get('labeled_SPD', None),
                'operational_SPD': json_result.get(p,
                                                   {}).get('fairness',
                                                           {}).get('operational_SPD', None),
                'impact_SPD': json_result.get(p,
                                              {}).get('fairness',
                                                      {}).get('impact_SPD', None),

                'training_TPR': json_result.get(p,
                                                {}).get('fairness',
                                                        {}).get('labeled_TPR', None),
                'training_FPR': json_result.get(p,
                                                {}).get('fairness',
                                                        {}).get('labeled_FPR', None),
                'training_PPV': json_result.get(p,
                                                {}).get('fairness',
                                                        {}).get('labeled_PPV', None),
                'training_PNV': json_result.get(p,
                                                {}).get('fairness',
                                                        {}).get('labeled_PNV', None),

                'training_equality': json_result.get(p,
                                                     {}).get('fairness',
                                                             {}).get('labeled_equality', None),
                'training_equity': json_result.get(p,
                                                   {}).get('fairness',
                                                           {}).get('labeled_equity', None),
                'operational_equality': json_result.get(p,
                                                        {}).get('fairness',
                                                                {}).get('operational_equality', None),
                'operational_equity': json_result.get(p,
                                                      {}).get('fairness',
                                                              {}).get('operational_equity', None),
                'impact_equality': json_result.get(p,
                                                   {}).get('fairness',
                                                           {}).get('impact_equality', None),
                'impact_equity': json_result.get(p,
                                                 {}).get('fairness',
                                                         {}).get('impact_equity', None),

                }
            audit_result[p_id]['impact_ratio'] = {}
            audit_result[p_id]['impact_ratio']['training'] = {
                'ref': 80,
                'df_result': model.labeled_results.get('impact_ratio',
                                                       {}).get(p, {}).get('df_result', None),
                'pass': model.labeled_results.get('impact_ratio',
                                                  {}).get(p, {}).get('pass', None),
                'total': model.labeled_results.get('impact_ratio',
                                                   {}).get(p, {}).get('total', None),
                'total_filter': model.labeled_results.get('impact_ratio',
                                                          {}).get(p, {}).get('total_filter', None),
                'unknown': model.labeled_results.get('impact_ratio',
                                                     {}).get(p, {}).get('unknown', None),
                }
            audit_result[p_id]['impact_ratio']['operational'] = {
                'ref': 80,
                'df_result': model.production_results.get('impact_ratio',
                                                          {}).get(p, {}).get('df_result', None),
                'pass': model.production_results.get('impact_ratio',
                                                     {}).get(p, {}).get('pass', None),
                'total': model.production_results.get('impact_ratio',
                                                      {}).get(p, {}).get('total', None),
                'total_filter': model.production_results.get('impact_ratio',
                                                             {}).get(p, {}).get('total_filter', None),
                'unknown': model.production_results.get('impact_ratio',
                                                        {}).get(p, {}).get('unknown', None),
                }
            audit_result[p_id]['impact_ratio']['impacted'] = {
                'ref': 80,
                'df_result': model.impacted_results.get('impact_ratio',
                                                        {}).get(p, {}).get('df_result', None),
                'pass': model.impacted_results.get('impact_ratio',
                                                   {}).get(p, {}).get('pass', None),
                'total': model.impacted_results.get('impact_ratio',
                                                    {}).get(p, {}).get('total', None),
                'total_filter': model.impacted_results.get('impact_ratio',
                                                           {}).get(p, {}).get('total_filter', None),
                'unknown': model.impacted_results.get('impact_ratio',
                                                      {}).get(p, {}).get('unknown', None),
                }

            audit_result[p_id]['performance'] = {
                'ref': 80,
                'inconsistency': json_result.get(p,
                                                 {}).get('drift', {}).get('drift', None),
                'poor_performance': json_result.get(p,
                                                    {}).get('performance', {}).get('poor_performance', None),
                'recall': json_result.get(p,
                                          {}).get('performance', {}).get('recall', None),
                'f1_score': json_result.get(p,
                                            {}).get('performance',  {}).get('f1_score', None),
                'accuracy': json_result.get(p,
                                            {}).get('performance', {}).get('accuracy', None),
                'precision': json_result.get(p,
                                             {}).get('performance', {}).get('precision', None),
                'TP': json_result.get(p,
                                      {}).get('performance', {}).get('TP', None),
                'FP': json_result.get(p,
                                      {}).get('performance', {}).get('FP', None),
                'TN': json_result.get(p,
                                      {}).get('performance', {}).get('TN', None),
                'FN': json_result.get(p,
                                      {}).get('performance', {}).get('FN', None),

                }
            files = ['training', 'operational', 'impact']

            evolution_params = {'score_first_last': {
                'key': 'benchmarking',
                'fields': ['training_data', 'operational_data', 'impact_data'],
                'overview': {'key0': 'share', 'key1': 'evol'},
                },
                'score_positives_first_last': {
                'key': 'benchmarking',
                'fields': ['training_positive', 'operational_positive', 'impact_positive'],
                'overview': {'key0': 'share_positives', 'key1': 'evol_positives'},
                },
                'score_DI_first_last': {
                'key': 'fairness',
                'fields': ['training_DI', 'operational_DI', 'impact_DI'],
                'overview': {'key0': 'share_DI', 'key1': 'evol_DI'},
                },
                'score_EQA_first_last': {
                'key': 'fairness',
                'fields': ['training_equality', 'operational_equality', 'impact_equality'],
                'overview': {'key0': 'share_EQA', 'key1': 'evol_EQA'},
                },
                'score_EQI_first_last': {
                'key': 'fairness',
                'fields': ['training_equity', 'operational_equity', 'impact_equity'],
                'overview': {'key0': 'share_EQI', 'key1': 'evol_EQI'},
                },
                'score_proxy_first_last': {
                'key': 'distribution',
                'fields': ['training_proxy', 'operational_proxy', 'impact_proxy'],
                'overview': {'key0': 'share_proxy', 'key1': 'evol_proxy'},
                },
                'score_label_first_last': {
                'key': 'distribution',
                'fields': ['training_label', 'operational_label', 'impact_label'],
                'overview': {'key0': 'share_label', 'key1': 'evol_label'},
                }}
            audit_result[p_id]['scoring_evolution'] = {}
            audit_result[p_id]['overview'] = {}
            for evolution in evolution_params.items():
                share_values = [audit_result[p_id][evolution[1]['key']][f] for f in evolution[1]['fields']]

                ref_evol = audit_result[p_id][evolution[1]['key']]['ref']
                no_none_share = [(v, f) for f, v in zip(files, share_values) if v is not None]
                if len(no_none_share) == 0:
                    share_first_last = [0] + [ref_evol]
                elif len(no_none_share) == 1:
                    share_first_last = [no_none_share[0][0]] + [ref_evol]
                else:
                    share_first_last = [no_none_share[0][0]] + [no_none_share[-1][0]]

                audit_result[p_id]['scoring_evolution'][evolution[0]] = scoring_evolution(share_first_last[0],
                                                                                          share_first_last[1],
                                                                                          ref_evol)
                audit_result[p_id]['overview'].update(overview(evolution[1]['overview']['key0'],
                                                               evolution[1]['overview']['key1'],
                                                               no_none_share,
                                                               ref_evol))
    result = {
        'detail': {
            'training_detail': model.labeled_results,
            'operational_detail': model.production_results,
            'impact_detail': model.impacted_results,
            'drift_detail': model.drift_results
        },
        'audit_result': audit_result
    }

    return result



import numpy as np
from .base_metric import BaseFairnessMetric
from ..utils.helpers import get_mask, binarize_predictions
import logging
logger = logging.getLogger(__name__)


class Performance(BaseFairnessMetric):
    """
    Metric D(A)-positive
    Calculates the distribution of a given group when label is positive.
    """

    def compute(self,
                input_data,
                sensitive_attrs=None,
                label_column: str = None,
                positive_outcome=[1],
                output_column: str = None):
        """
        Parameters
        ----------
       :param input_data: Dataset which should contain the columns in param sensitive_attrs.
       :param sensitive_attrs: The sensitive attributes (e.g., 'gender', 'race') relevant to fairness.
       :param label_column: Name of the column containing the target.
       :param output_column: Name of the column containing the prediction / classification.
        Returns
        ----------
       :return: A dictionary containing the results for each group
        """
        self.validate_parameters(
            input_data=input_data,
            sensitive_attrs=sensitive_attrs,
            label_column=label_column,
            positive_outcome=positive_outcome,
            output_column=output_column
        )
        np.random.seed(123)
        # Verify training data
        # verify_training(csv_path)
        if not input_data[label_column].isin([0, 1]).all():
            # Convert to binary
            input_data[label_column] = binarize_predictions(input_data[label_column].values)
        if not input_data[output_column].isin([0, 1]).all():
            # Convert to binary
            input_data[output_column] = binarize_predictions(input_data[output_column].values)

        json_groups = sensitive_attrs

        result_list = {}
        TP = ((input_data[label_column].isin(positive_outcome)) &
              (input_data[output_column].isin(positive_outcome))).sum().item()  # True Positives
        FP = ((~input_data[label_column].isin(positive_outcome)) &
              (input_data[output_column].isin(positive_outcome))).sum().item()  # False Positives
        TN = ((~input_data[label_column].isin(positive_outcome)) &
              (~input_data[output_column].isin(positive_outcome))).sum().item()  # True Negatives
        FN = ((input_data[label_column].isin(positive_outcome)) &
              (~input_data[output_column].isin(positive_outcome))).sum().item()  # False Negatives
        accuracy = np.round((TP + TN) / (TP + TN + FP + FN), 4).item()
        precision = np.round(TP / (TP + FP), 4).item() if (TP + FP) > 0 else 0
        recall = np.round(TP / (TP + FN), 4).item() if (TP + FN) > 0 else 0
        f1 = np.round(2 * (precision * recall) / (precision + recall), 4).item() if (precision + recall) > 0 else 0
        largest_class_frequency = np.round(input_data[label_column].value_counts().max() / len(input_data), 4).item()
        poor_performance = np.round(accuracy / largest_class_frequency, 4).item()
        normalized_risk = self.normalize_value(poor_performance)
        result_list.update({'overall': {
                                    'accuracy': np.round(accuracy * 100, 4).item(),
                                    'precision': np.round(precision * 100, 4).item(),
                                    'recall': np.round(recall * 100, 4).item(),
                                    'f1': np.round(f1 * 100, 4).item(),
                                    'lcf': largest_class_frequency,
                                    'TP': int(TP),
                                    'FP': int(FP),
                                    'TN': int(TN),
                                    'FN': int(FN),
                                    'poor_performance': poor_performance,
                                    'normalized_risk': normalized_risk,
                                    'bias_level': self.get_bias_level(normalized_risk)}})
        for item in json_groups.items():

            group = item[0]
            try:
                if item[1]['type'] == 'simple':
                    filters = item[1]['columns']
                else:
                    filters = np.concat([json_groups[c]['columns'] for c in item[1]['groups']]).tolist()

                mask_privileged, mask_underprivileged = get_mask(input_data, filters)

                data = input_data[mask_underprivileged].copy()
                TP = ((data[label_column].isin(positive_outcome)) &
                      (data[output_column].isin(positive_outcome))).sum()  # True Positives
                FP = ((~data[label_column].isin(positive_outcome)) &
                      (data[output_column].isin(positive_outcome))).sum()  # False Positives
                TN = ((~data[label_column].isin(positive_outcome)) &
                      (~data[output_column].isin(positive_outcome))).sum()  # True Negatives
                FN = ((data[label_column].isin(positive_outcome)) &
                      (~data[output_column].isin(positive_outcome))).sum()  # False Negatives
                accuracy = np.round((TP + TN) / (TP + TN + FP + FN), 4).item()
                precision = np.round(TP / (TP + FP), 4).item() if (TP + FP) > 0 else 0
                recall = np.round(TP / (TP + FN), 4).item() if (TP + FN) > 0 else 0
                f1 = 0
                if (precision + recall) > 0:
                    f1 = np.round(2 * (precision * recall) / (precision + recall), 4).item()

                largest_class_frequency = np.round(data['outcome'].value_counts().max() / len(data), 4).item()
                poor_performance = np.round(accuracy / largest_class_frequency, 4).item()
                normalized_risk = self.normalize_value(poor_performance)
                result_list.update({group: {
                                            'accuracy': np.round(accuracy * 100, 4).item(),
                                            'precision': np.round(precision * 100, 4).item(),
                                            'recall': np.round(recall * 100, 4).item(),
                                            'f1': np.round(f1 * 100, 4).item(),
                                            'lcf': largest_class_frequency,
                                            'TP': int(TP),
                                            'FP': int(FP),
                                            'TN': int(TN),
                                            'FN': int(FN),
                                            'poor_performance': np.round(poor_performance, 4).item(),
                                            'normalized_risk': np.round(normalized_risk, 4).item(),
                                            'bias_level': self.get_bias_level(normalized_risk)}})
            except KeyError:
                logger.error(f"Group no present in data: '{group}'")
                # Captura cualquier ValueError lanzado por get_benchmarking
                result_list.update({group: {
                                            'accuracy': 0,
                                            'precision': 0,
                                            'recall': 0,
                                            'f1': 0,
                                            'lcf': 0,
                                            'TP': int(TP),
                                            'FP': int(FP),
                                            'TN': int(TN),
                                            'FN': int(FN),
                                            'poor_performance': 0,
                                            'normalized_risk': None,
                                            'bias_level': self.get_bias_level(0),
                                            'error': 'group no present in data.'}})
        logger.info(f"Completed: '{self.__str__()}'")
        return result_list

    def normalize_value(self, value):
        scaled_accuracy = value
        if scaled_accuracy < 1:
            return scaled_accuracy * 60 / 1
        elif 1.0 <= scaled_accuracy < 1.3:
            return 60 + (scaled_accuracy - 1) * 20 / 0.3
        elif 1.3 <= scaled_accuracy <= 2:
            return 80 + (scaled_accuracy - 1.3) * 20 / 0.7
        else:
            return 100



from eticas.audit.base_audit import BaseAudit
import unittest


class TestBaseAudit(BaseAudit):

    def run_audit(self, dataset_path, label_column, output_column, positive_output):
        return super().run_audit(dataset_path, label_column, output_column, positive_output)


class TestBaseFairnessMetric(unittest.TestCase):

    def setUp(self):
        # Create an instance of the dummy metric
        self.audit = TestBaseAudit(None)
        # Create a simple DataFrame for testing

    def test_audit(self):

        # Attempting to call compute should raise NotImplementedError.
        with self.assertRaises(NotImplementedError) as context:
            self.audit.run_audit(None, None, None, None)
        # Optional: Assert that the exception message is as expected.
        self.assertEqual(str(context.exception), "Not implemented.")

    def test_str_method(self):
        # Test that the __str__ method returns the class name.
        self.assertEqual(str(self.audit), "TestBaseAudit", "The __str__ method should return the class name.")


if __name__ == '__main__':
    unittest.main()



import numpy as np


def binarize_predictions(scores):
    """
    Converts a list/array of continuous scores into binary predictions (0/1)
    using the mean.
    """
    mean_value = np.mean(scores)
    return np.where(scores > mean_value, 1, 0)


def get_mask(input_data, filters):
    """
    Get boolean array to filter data base on filters: Name of column and Conditions.
    {
        "name": "ethnicity",
        "privileged": [1]
    }
    """
    mask_privileged = np.ones(input_data.shape[0], dtype=bool)
    mask_underprivileged = np.ones(input_data.shape[0], dtype=bool)
    for filter in filters:

        if 'privileged' in filter:
            mask_underprivileged = mask_underprivileged & (
                ~input_data[filter['name']].isin(filter['privileged']))
            mask_privileged = mask_privileged & (
                input_data[filter['name']].isin(filter['privileged']))
        else:
            mask_underprivileged = mask_underprivileged & (
                input_data[filter['name']].isin(filter['underprivileged']))
            mask_privileged = mask_privileged & (
                ~input_data[filter['name']].isin(filter['underprivileged']))

    return mask_privileged, mask_underprivileged



"""
production_audit.py
===================

Implements an audit for the production (or testing) phase of a model.
"""


from .base_audit import BaseAudit
from ..data.loaders import load_dataset

from ..metrics.tdx_inconsistency import Tdx_inconsistency
import logging
logger = logging.getLogger(__name__)


class DriftAudit(BaseAudit):
    """
    Drift Audit.
    """
    def run_audit(self,
                  dataset_path_dev: str,
                  output_column_dev: str,
                  positive_output_dev: list,
                  dataset_path_prod: str,
                  output_column_prod: str,
                  positive_output_prod: list):
        """
        Performs checks related to the unlabeled / production  dataset and process.

       :param dataset_path: path to dataset.
       :param output_column: Name of the column containing the prediction / classification.
       :param positive_output: Values of the column_output consider as positive.

        Returns
        -------
       :return: dict. The result of the unlabeled audit.
        """
        # Example: Check for data imbalance, missing values, etc.
        logger.info(f"Running drift audit on model '{self.model.model_name}'")
        input_data_dev = load_dataset(dataset_path_dev)
        input_data_dev = input_data_dev.dropna()
        logger.info(f"DEV data loaded '{input_data_dev.shape}'")
        if input_data_dev.shape[0] == 0:
            raise ValueError("DEV data shape is 0.")

        input_data_prod = load_dataset(dataset_path_prod)
        input_data_prod = input_data_prod.dropna()
        logger.info(f"PROD data loaded '{input_data_prod.shape}'")
        if input_data_prod.shape[0] == 0:
            raise ValueError("PROD data shape is 0.")

        drift_log = {}
        drift_log.update({'tdx_inconsistency': Tdx_inconsistency().compute(self.model.sensitive_attributes,
                                                                           self.model.features,
                                                                           input_data_dev,
                                                                           output_column_dev,
                                                                           positive_output_dev,
                                                                           input_data_prod,
                                                                           output_column_prod,
                                                                           positive_output_prod)})
        return drift_log



# run test python -m unittest discover tests

import logging
import unittest
from eticas.model.ml_model import MLModel
from eticas.audit.labeled_audit import LabeledAudit
from eticas.audit.unlabeled_audit import UnlabeledAudit
from eticas.audit.drift_audit import DriftAudit
from eticas.utils.api import upload_json_audit
import pandas as pd
import pandas.testing as pdt
import json
logger = logging.getLogger(__name__)

sensitive_attributes = {'sex': {'columns': [
                                            {
                                                "name": "sex",
                                                "underprivileged": [2]
                                            }
                                        ],
                                'type': 'simple'},
                        'ethnicity': {'columns': [
                                                 {
                                                     "name": "ethnicity",
                                                     "privileged": [1]
                                                 }
                                        ],
                                      'type': 'simple'},
                        'age': {'columns': [
                                            {
                                                 "name": "age",
                                                 "privileged": [3, 4]
                                            }
                                        ],
                                'type': 'simple'},
                        'error': {'columns': [
                            {
                                "name": "error",
                                  "privileged": [3, 4]
                                  }
                                  ],
                                  'type': 'simple'},
                        'sex_ethnicity': {'groups': [
                                                    "sex", "ethnicity"
                                                    ],
                                          'type': 'complex'}}
distribution_ref = {'sex': 40,
                    'ethnicity': 40,
                    'age': 40,
                    'sex_ethnicity': 40
                    }
sensitive_attribute2 = {'sex': {'columns': [
                                            {
                                                "name": "sex",
                                                "underprivileged": [2]
                                            }
                                        ],
                                'type': 'simple'},
                        'ethnicity': {'columns': [
                                                 {
                                                     "name": "ethnicity",
                                                     "privileged": [1]
                                                 }
                                        ],
                                      'type': 'simple'},
                        'age': {'columns': [
                                            {
                                                 "name": "age",
                                                 "privileged": [3, 4]
                                            }
                                        ],
                                'type': 'simple'},
                        'no_columns': {'columns': [
                            {
                                "name": "no_column",
                                "privileged": [3, 4]
                                  }
                                  ],
                                  'type': 'simple'},
                        'sex_ethnicity': {'groups': [
                                                    "sex", "ethnicity"
                                                    ],
                                          'type': 'complex'}}
logging.basicConfig(
    level=logging.ERROR,
    format='[%(levelname)s] %(name)s - %(message)s'
)


class TestMetrics(unittest.TestCase):

    def setUp(self):
        self.model = MLModel(
            model_name="ML Testing Regression",
            description="A logistic regression model to illustrate audits",
            country="USA",
            state="CA",
            sensitive_attributes=sensitive_attributes,
            distribution_ref=distribution_ref,
            features=["feature_0", "feature_1", "feature_2"]
        )
        self.model.run_labeled_audit(dataset_path='files/example_training_binary_2.csv',
                                     label_column='outcome', output_column='predicted_outcome', positive_output=[1])
        self.model.run_production_audit(dataset_path='files/example_operational_binary_2.csv',
                                        output_column='predicted_outcome', positive_output=[1])
        self.model.run_impacted_audit(dataset_path='files/example_impact_binary_2.csv',
                                      output_column='recorded_outcome', positive_output=[1])
        self.model.run_drift_audit(dataset_path_dev='files/example_training_binary_2.csv',
                                   output_column_dev='outcome', positive_output_dev=[1],
                                   dataset_path_prod='files/example_operational_binary_2.csv',
                                   output_column_prod='predicted_outcome', positive_output_prod=[1])

    def test_df_norm(self):
        expected = pd.read_pickle('tests/ml_files/result_df_norm.pickle')
        result = self.model.df_results(norm_values=True).reset_index()
        pdt.assert_frame_equal(result, expected)

    def test_df(self):
        expected = pd.read_pickle('tests/ml_files/result_df.pickle')
        result = self.model.df_results(norm_values=False).reset_index()
        pdt.assert_frame_equal(result, expected)

    def test_json_norm(self):
        with open('tests/ml_files/result_json_norm.json', 'r', encoding='utf-8') as f:
            expected = json.load(f)
        result = self.model.json_results(norm_values=True)
        self.assertEqual(result, expected)

    def test_json(self):
        with open('tests/ml_files/result_json.json', 'r', encoding='utf-8') as f:
            expected = json.load(f)
        result = self.model.json_results(norm_values=False)
        self.assertEqual(result, expected)

    def test_json_labeled(self):
        with open('tests/ml_files/result_json_labeled.json', 'r', encoding='utf-8') as f:
            expected = json.load(f)
        result = self.model.labeled_results
        self.assertEqual(result, expected)

    def test_json_production(self):
        with open('tests/ml_files/result_json_production.json', 'r', encoding='utf-8') as f:
            expected = json.load(f)
        result = self.model.production_results
        self.assertEqual(result, expected)

    def test_json_impacted(self):
        with open('tests/ml_files/result_json_impacted.json', 'r', encoding='utf-8') as f:
            expected = json.load(f)
        result = self.model.impacted_results
        self.assertEqual(result, expected)

    def test_json_upload(self):
        result = upload_json_audit(model=self.model)
        self.assertIsInstance(result, dict)
        model_label = MLModel(
            model_name="ML Testing Regression",
            description="A logistic regression model to illustrate audits",
            country="USA",
            state="CA",
            sensitive_attributes=sensitive_attribute2,
            features=["feature_0"]
        )
        model_label.run_labeled_audit(dataset_path='files/example_training_scoring.csv',
                                      label_column='outcome', output_column='predicted_outcome', positive_output=[1])
        model_label.run_impacted_audit(dataset_path='files/example_impact_scoring.csv',
                                       output_column='recorded_outcome', positive_output=[1])
        result = upload_json_audit(model=model_label)
        self.assertIsInstance(result, dict)
        model_prod = MLModel(
            model_name="ML Testing Regression",
            description="A logistic regression model to illustrate audits",
            country="USA",
            state="CA",
            sensitive_attributes=sensitive_attributes,
            features=["feature_0", "feature_1", "feature_2"]
        )
        model_prod.run_production_audit(dataset_path='files/example_operational_binary_2.csv',
                                        output_column='predicted_outcome', positive_output=[1])
        result = upload_json_audit(model=model_prod)
        self.assertIsInstance(result, dict)
        model_none = MLModel(
            model_name="ML Testing Regression",
            description="A logistic regression model to illustrate audits",
            country="USA",
            state="CA",
            sensitive_attributes=sensitive_attributes,
            features=["feature_0", "feature_1", "feature_2"]
        )
        result = upload_json_audit(model=model_none)
        self.assertIsInstance(result, dict)

    def test_test_drift_dev_shape(self):
        with self.assertRaises(ValueError) as context:
            DriftAudit(self.model).run_audit('tests/ml_files/zero_shape.csv', None, None, None, None, None)
        self.assertIn("DEV data shape is 0.", str(context.exception))

    def test_test_drift_prod_shape(self):
        with self.assertRaises(ValueError) as context:
            DriftAudit(self.model).run_audit('files/example_training_binary_2.csv', None, None,
                                             'tests/ml_files/zero_shape.csv', None, None)
        self.assertIn("PROD data shape is 0.", str(context.exception))

    def test_test_labeled_shape(self):
        with self.assertRaises(ValueError) as context:
            LabeledAudit(self.model).run_audit('tests/ml_files/zero_shape.csv', None, None)
        self.assertIn("labeled dataset shape is 0.", str(context.exception))

    def test_test_unlabeled_shape(self):
        with self.assertRaises(ValueError) as context:
            UnlabeledAudit(self.model).run_audit('tests/ml_files/zero_shape.csv', None, None)
        self.assertIn("production dataset shape is 0.", str(context.exception))

    def test_str_method(self):
        # Test that the __str__ method returns the class name.
        self.assertEqual(str(self.model), "MLModel(ML Testing Regression)",
                         "The __str__ method should return the class name.")

    def test_disparate_impact(self):

        df_labeled = self.model.impact_ratio('labeled', 'sex')
        self.assertIsInstance(df_labeled, pd.DataFrame)

        df_production = self.model.impact_ratio('production', 'sex')
        self.assertIsInstance(df_production, pd.DataFrame)

        df_impacted = self.model.impact_ratio('impacted', 'sex')
        self.assertIsInstance(df_impacted, pd.DataFrame)

        with self.assertRaises(ValueError) as context:
            self.model.impact_ratio('labeled', 'invalid')
        self.assertEqual(str(context.exception), "You must provide a correct attribute.")

        with self.assertRaises(ValueError) as context:
            self.model.impact_ratio('production', 'invalid')
        self.assertEqual(str(context.exception), "You must provide a correct attribute.")

        with self.assertRaises(ValueError) as context:
            self.model.impact_ratio('impacted', 'invalid')
        self.assertEqual(str(context.exception), "You must provide a correct attribute.")

        with self.assertRaises(ValueError) as context:
            self.model.impact_ratio('unknown', 'XXXX')
        self.assertEqual(str(context.exception), "You must provide a correct stage.")


if __name__ == '__main__':
    unittest.main()



import numpy as np
from .base_metric import BaseFairnessMetric
from ..utils.helpers import get_mask
import logging
logger = logging.getLogger(__name__)


class Da_inconsistency(BaseFairnessMetric):
    """
    Metric D(A)-inconsistency
    Calculates the distribution of a given group
    """

    def compute(self,
                input_data,
                sensitive_attrs=None):
        """
        Parameters
        ----------
       :param input_data: Dataset which should contain the columns in param sensitive_attrs.
       :param sensitive_attrs: The sensitive attributes (e.g., 'gender', 'race') relevant to fairness.
        Returns
        ----------
       :return: A dictionary containing the results for each group
        """
        self.validate_parameters(
            input_data=input_data,
            sensitive_attrs=sensitive_attrs
        )
        json_groups = sensitive_attrs
        result_list = {}

        for item in json_groups.items():
            group = item[0]

            if item[1]['type'] == 'simple':
                filters = item[1]['columns']
            else:
                filters = np.concat([json_groups[c]['columns'] for c in item[1]['groups']]).tolist()

            try:
                mask_privileged, mask_underprivileged = get_mask(input_data, filters)
                data = np.round((mask_underprivileged.sum() / mask_underprivileged.shape[0])*100, 1).item()

                result_list.update({group: {'data': data}})
            except KeyError:
                # Captura cualquier ValueError lanzado por get_benchmarking

                logger.error(f"Group no present in data: '{group}'")
                result_list.update({group: {'data': None,
                                            'error': 'group no present in data.'}})
        logger.info(f"Completed: '{self.__str__()}'")
        return result_list



import os
import logging
import pandas as pd

# Basic logger configuration (customize it to suit your needs)
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)
logger = logging.getLogger(__name__)


def load_dataset(file_path: str) -> pd.DataFrame:
    """
    Loads a dataset from a .pkl, .parquet, or .csv file.

    Parameters
    ----------
    file_path: str
        Full path to the file to be read.

    Returns
    -------
    pd.DataFrame
        A pandas DataFrame containing the file data.

    Raises
    ------
    FileNotFoundError
        If the file does not exist.
    ValueError
        If the file extension is not supported.
    Exception
        If an error occurs while loading the file.
    """
    # 1. Check if the file exists
    if not os.path.exists(file_path):
        error_msg = f"The file '{file_path}' does not exist."
        logger.error(error_msg)
        raise FileNotFoundError(error_msg)

    # 2. Identify the file extension
    _, extension = os.path.splitext(file_path)
    extension = extension.lower()

    # 3. Load the data based on the extension
    if (extension == '.pkl') | (extension == '.pickle'):
        df = pd.read_pickle(file_path)
        logger.info(f"Dataset successfully loaded from {file_path} (Pickle).")
    elif extension == '.parquet':
        df = pd.read_parquet(file_path)
        logger.info(f"Dataset successfully loaded from {file_path} (Parquet).")
    elif extension == '.csv':
        df = pd.read_csv(file_path)
        logger.info(f"Dataset successfully loaded from {file_path} (CSV).")
    else:
        error_msg = (f"Extension '{extension}' is not supported. "
                     f"Only .pkl, .parquet, or .csv files are allowed.")
        logger.error(error_msg)
        raise ValueError(error_msg)

    return df



"""
base_model.py
=============

Defines a base class for models in the eticas library.
"""

from abc import ABC


class BaseModel(ABC):
    """
    Base class for a model in the eticas library.
    """

    def __init__(
        self,
        model_name: str,
        description: str = None,
        country: str = None,
        state: str = None,
        sensitive_attributes: dict = None,
        distribution_ref: dict = None,
        features: list = None
    ):
        """
        Initialize basic model metadata.

        Parameters
        ----------
        :param model_name : str
            Name of the model (e.g., 'Logistic Regression', 'XGBoost', etc.).
        :param description : str, optional
            A brief description of the model (version, purpose, etc.).
        :param country : str, optional
            The country where the model is primarily used or deployed.
        :param state : str, optional
            The state or region where the model is primarily used.
        :param sensitive_attributes : dict, optional
            The sensitive attributes (e.g., 'gender', 'race') relevant to fairness.
                The column in dataset.
                Under-privileged values.
                {'gender' : {'columns' : [
                                            {
                                            "name": "sex",
                                            "underprivileged": [2]
                                            }
                                        ],
                             'type' : 'simple'},
                 'ethnicity' : {'columns' : [
                                                        {
                                                        "name": "ethnicity",
                                                        "privileged": [1]
                                                        }
                                                    ],
                                         'type' : 'simple'}
                 'gender_ethnicity' : {'groups' : ["gender","ethnicity"],
                                                'type' : 'complex'}}
        :param distribution_ref : dict Expected distribution for underprivileged group in production.
        :param features : list, optional
            The features used by the model (column names in a dataset).
        """
        self.model_name = model_name
        self.description = description
        self.country = country
        self.state = state
        self.sensitive_attributes = sensitive_attributes
        self.features = features
        self.labeled_results = {}
        self.production_results = {}
        self.impacted_results = {}
        self.drift_results = {}
        if distribution_ref is None:
            self.distribution_ref = {}
        else:
            self.distribution_ref = distribution_ref

    def __str__(self):
        return f"{self.__class__.__name__}({self.model_name})"



"""
base_audit.py
=============

Defines the abstract base class for model audits.
"""

from abc import ABC, abstractmethod


class BaseAudit(ABC):
    """
    Abstract base class for auditing a model.
    """

    def __init__(self, model):
        """
        Parameters
        ----------
        model : BaseModel
            The model instance to be audited.
        """
        self.model = model

    @abstractmethod
    def run_audit(self,
                  dataset_path: str,
                  label_column: str,
                  output_column: str = None,
                  positive_output: list = None):
        """
        :param dataset_path : path to training dataset.
        :param label_column: Name of the column containing the target.
        :param output_column: Name of the column containing the prediction / classification.
        :param positive_output: Values of the column_output consider as positive.


        Returns
        -------
        :return: dict. The result of the training audit.
        """
        raise NotImplementedError("Not implemented.")

    def __str__(self):
        return self.__class__.__name__



import os
import unittest
from unittest.mock import patch, MagicMock
import requests
import pandas as pd

# Importa las funciones a testear desde tu módulo
# coverage run -m unittest discover
# coverage html
from eticas.utils.api import get_audit, get_departments, get_models, get_audits, upload_audit
from eticas.utils.api import scoring_evolution, bias_direction, overview


class TestAPIMethods(unittest.TestCase):

    def setUp(self):
        os.environ["ITACA_API_TOKEN"] = "dummy_token"
        os.environ["ITACA_BASE_URL"] = "https://itaca.eticas.ai/api/v1/"

    @patch("eticas.utils.api.requests.get")
    def test_get_audit_success(self, mock_get):
        expected_response = {"id": 2289, "audit": "dummy audit"}
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = expected_response
        mock_get.return_value = mock_response

        result = get_audit(audit_id=2289)
        self.assertEqual(result, expected_response)
        mock_get.assert_called_once()

    @patch("eticas.utils.api.requests.get")
    def test_get_audit_failure(self, mock_get):
        error_text = "Not Found"
        mock_response = MagicMock()
        mock_response.status_code = 404
        mock_response.text = error_text
        mock_get.return_value = mock_response

        with self.assertRaises(requests.HTTPError):
            get_audit(audit_id=999)

    @patch("eticas.utils.api.requests.get")
    def test_get_departments_success(self, mock_get):
        expected_data = [{"id": 1, "name": "HR"}, {"id": 2, "name": "IT"}]
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = expected_data
        mock_get.return_value = mock_response

        df = get_departments()
        self.assertIsInstance(df, pd.DataFrame)
        self.assertEqual(len(df), len(expected_data))

    @patch("eticas.utils.api.requests.get")
    def test_get_model_success(self, mock_get):
        expected_data = [{"id": 10, "name": "Model A"}]
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = expected_data
        mock_get.return_value = mock_response

        df = get_models(department=216)
        self.assertIsInstance(df, pd.DataFrame)
        self.assertEqual(len(df), len(expected_data))

    @patch("eticas.utils.api.requests.get")
    def test_get_audits_success(self, mock_get):
        expected_results = {"results": [{"id": 100, "name": "Audit 1"}, {"id": 101, "name": "Audit 2"}]}
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = expected_results
        mock_get.return_value = mock_response

        df = get_audits(model=263)
        self.assertIsInstance(df, pd.DataFrame)
        self.assertEqual(len(df), len(expected_results["results"]))

    @patch.dict(os.environ, {'ITACA_BASE_URL': "https://itaca.eticas.ai/api/v1/"}, clear=True)
    def test_missing_api_key(self):
        with self.assertRaises(ValueError) as context:
            get_audit(audit_id=2289)
        self.assertEqual(str(context.exception), "❌ 'ITACA_API_TOKEN' NO DEFINED.")

        with self.assertRaises(ValueError) as context:
            get_departments()
        self.assertEqual(str(context.exception), "❌ 'ITACA_API_TOKEN' NO DEFINED.")

        with self.assertRaises(ValueError) as context:
            get_models(department=216)
        self.assertEqual(str(context.exception), "❌ 'ITACA_API_TOKEN' NO DEFINED.")

        with self.assertRaises(ValueError) as context:
            get_audits(model=263)
        self.assertEqual(str(context.exception), "❌ 'ITACA_API_TOKEN' NO DEFINED.")

    @patch.dict(os.environ, {}, clear=True)
    def test_missing_base_url(self):
        with self.assertRaises(ValueError) as context:
            get_audit(audit_id=2289)
        self.assertEqual(str(context.exception), "❌ 'ITACA_BASE_URL' NO DEFINED.")

        with self.assertRaises(ValueError) as context:
            get_departments()
        self.assertEqual(str(context.exception), "❌ 'ITACA_BASE_URL' NO DEFINED.")

        with self.assertRaises(ValueError) as context:
            get_models(department=216)
        self.assertEqual(str(context.exception), "❌ 'ITACA_BASE_URL' NO DEFINED.")

        with self.assertRaises(ValueError) as context:
            get_audits(model=263)
        self.assertEqual(str(context.exception), "❌ 'ITACA_BASE_URL' NO DEFINED.")

    @patch("eticas.utils.api.requests.get")
    def test_api_error_responses(self, mock_get):
        error_text = "Internal Server Error"
        mock_response = MagicMock()
        mock_response.status_code = 500
        mock_response.text = error_text
        mock_get.return_value = mock_response

        with self.assertRaises(requests.HTTPError) as context:
            get_audit(audit_id=2289)
        self.assertIn("500", str(context.exception))

        with self.assertRaises(requests.HTTPError) as context:
            get_departments()
        self.assertIn("500", str(context.exception))

        with self.assertRaises(requests.HTTPError) as context:
            get_models(department=216)
        self.assertIn("500", str(context.exception))

        with self.assertRaises(requests.HTTPError) as context:
            get_audits(model=263)
        self.assertIn("500", str(context.exception))

    @patch.dict(os.environ, {'ITACA_BASE_URL': "https://itaca.eticas.ai/api/v1/"}, clear=True)
    def test_upload_audit_missing_api_key(self):
        with self.assertRaises(ValueError) as context:
            upload_audit(department_id=1, model_id=100, model={"dummy": "data"})
        self.assertEqual(str(context.exception), "❌ 'ITACA_API_TOKEN' NO DEFINED.")

    @patch.dict(os.environ, {'ITACA_API_TOKEN': "DUMMY"}, clear=True)
    def test_upload_audit_missing_base_url(self):
        with self.assertRaises(ValueError) as context:
            upload_audit(department_id=1, model_id=100, model={"dummy": "data"})
        self.assertEqual(str(context.exception), "❌ 'ITACA_BASE_URL' NO DEFINED.")

    @patch("eticas.utils.api.get_departments")
    def test_upload_audit_invalid_department_empty(self, mock_get_departments):
        mock_get_departments.return_value = pd.DataFrame()
        with self.assertRaises(ValueError) as context:
            upload_audit(department_id=1, model_id=100, model={"dummy": "data"})
        self.assertEqual(str(context.exception), "Deparment ID does not exist..")

    @patch("eticas.utils.api.get_departments")
    def test_upload_audit_invalid_department_not_in_list(self, mock_get_departments):
        df_dept = pd.DataFrame({"id": [10, 20, 30]})
        mock_get_departments.return_value = df_dept
        with self.assertRaises(ValueError) as context:
            upload_audit(department_id=1, model_id=100, model={"dummy": "data"})
        self.assertEqual(str(context.exception), "Deparment ID does not exist..")

    @patch("eticas.utils.api.get_models")
    @patch("eticas.utils.api.get_departments")
    def test_upload_audit_invalid_model_empty(self, mock_get_departments, mock_get_models):
        df_dept = pd.DataFrame({"id": [1, 2, 3]})
        mock_get_departments.return_value = df_dept
        mock_get_models.return_value = pd.DataFrame()
        with self.assertRaises(ValueError) as context:
            upload_audit(department_id=1, model_id=100, model={"dummy": "data"})
        self.assertEqual(str(context.exception), "Model ID does not exist..")

    @patch("eticas.utils.api.get_models")
    @patch("eticas.utils.api.get_departments")
    def test_upload_audit_invalid_model_not_in_list(self, mock_get_departments, mock_get_models):
        df_dept = pd.DataFrame({"id": [1, 2, 3]})
        mock_get_departments.return_value = df_dept
        df_models = pd.DataFrame({"id": [10, 20, 30]})
        mock_get_models.return_value = df_models
        with self.assertRaises(ValueError) as context:
            upload_audit(department_id=1, model_id=100, model={"dummy": "data"})
        self.assertEqual(str(context.exception), "Model ID does not exist..")

    @patch("eticas.utils.api.requests.post")
    @patch("eticas.utils.api.upload_json_audit")
    @patch("eticas.utils.api.get_models")
    @patch("eticas.utils.api.get_departments")
    def test_upload_audit_success(self, mock_get_departments, mock_get_models, mock_upload_json_audit, mock_post):
        df_dept = pd.DataFrame({"id": [1, 2, 3]})
        mock_get_departments.return_value = df_dept
        df_models = pd.DataFrame({"id": [100, 200, 300]})
        mock_get_models.return_value = df_models
        audit_metrics = {"metric": "value"}
        mock_upload_json_audit.return_value = audit_metrics
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_post.return_value = mock_response

        department_id = 1
        model_id = 100
        model_input = {"dummy": "data"}
        status_code = upload_audit(department_id=department_id, model_id=model_id, model=model_input)

        self.assertEqual(status_code, 200)

    def test_bias_direction(self):
        self.assertEqual(bias_direction(1), "Correct-representation")
        self.assertEqual(bias_direction(0.5), "Under-representation")
        self.assertEqual(bias_direction(2), "Over-representation")

    def test_overview_single_element_equal(self):
        key0 = "share"
        key1 = "delta"
        no_none_share = [(10, "groupA")]
        ref_evol = 10
        expected = {
            "share_groupA_difference": 0,
            "share_groupA_evol": "equal to"
        }
        result = overview(key0, key1, no_none_share, ref_evol)
        self.assertEqual(result, expected)

    def test_overview_single_element_above(self):
        key0 = "share"
        key1 = "delta"
        no_none_share = [(15, "groupA")]
        ref_evol = 10
        expected = {
            "share_groupA_difference": 5,
            "share_groupA_evol": "above"
        }
        result = overview(key0, key1, no_none_share, ref_evol)
        self.assertEqual(result, expected)

    def test_overview_single_element_below(self):
        key0 = "share"
        key1 = "delta"
        no_none_share = [(8, "groupA")]
        ref_evol = 10
        expected = {
            "share_groupA_difference": 2,
            "share_groupA_evol": "below"
        }
        result = overview(key0, key1, no_none_share, ref_evol)
        self.assertEqual(result, expected)

    def test_overview_two_elements(self):
        key0 = "share"
        key1 = "delta"
        no_none_share = [(8, "A"), (12, "B")]
        ref_evol = 10
        expected = {
            "share_A_difference": 2,
            "share_A_evol": "below",
            "share_B_difference": 2,
            "share_B_evol": "above",
            "delta_A_B": "positive"
        }
        result = overview(key0, key1, no_none_share, ref_evol)
        self.assertEqual(result, expected)

    def test_overview_three_elements(self):
        key0 = "k0"
        key1 = "k1"
        no_none_share = [(10, "X"), (10, "Y"), (5, "Z")]
        ref_evol = 10
        expected = {
            "k0_X_difference": 0,
            "k0_X_evol": "equal to",
            "k0_Y_difference": 0,
            "k0_Y_evol": "equal to",
            "k0_Z_difference": 5,
            "k0_Z_evol": "below",
            "k1_X_Y": "neutral",
            "k1_Y_Z": "negative"
        }
        result = overview(key0, key1, no_none_share, ref_evol)
        self.assertEqual(result, expected)

    def test_scoring_evolution_above_ref(self):
        first_share = 5
        last_share = 15
        ref_share = 10
        result = scoring_evolution(first_share, last_share, ref_share)
        self.assertEqual(result, 100)

    def test_scoring_evolution_below_first(self):
        first_share = 10
        last_share = 5
        ref_share = 8
        result = scoring_evolution(first_share, last_share, ref_share)
        self.assertEqual(result, 0)

    def test_scoring_evolution_normalized(self):
        first_share = 10
        last_share = 12
        ref_share = 20
        result = scoring_evolution(first_share, last_share, ref_share)
        self.assertAlmostEqual(result, 12.0, places=4)


if __name__ == "__main__":
    unittest.main()



"""
audit
=====

This package manages audits for a given model. An audit can consist of different types:
- Training
- Production (or Testing)
- Recorded (or Impacted)

Each audit class captures logic and data relevant to that auditing phase.
"""



import numpy as np
from .base_metric import BaseFairnessMetric
from ..utils.helpers import get_mask, binarize_predictions
import logging
logger = logging.getLogger(__name__)


class D_parity(BaseFairnessMetric):
    """
    Metric D- Parity

    Distribution of the predicted labels differs between protected and reference groups in the population (DI).
    """

    def compute(self,
                input_data,
                sensitive_attrs=None,
                label_column: str = None,
                positive_output: list = [1]):
        """
        Parameters
        ----------
       :param input_data: Dataset which should contain the columns in param sensitive_attrs.
       :param sensitive_attrs: The sensitive attributes (e.g., 'gender', 'race') relevant to fairness.
       :param label_column: Name of the column containing the target.
       :param positive_output: Values of the column_output consider as positive.
        Returns
        ----------
       :return: A dictionary containing the results for each group
        """

        self.validate_parameters(
            input_data=input_data,
            sensitive_attrs=sensitive_attrs,
            label_column=label_column,
            positive_output=positive_output
        )

        if not input_data[label_column].isin([0, 1]).all():
            input_data[label_column] = binarize_predictions(input_data[label_column].values)
        json_groups = sensitive_attrs
        result_list = {}

        for item in json_groups.items():
            group = item[0]
            if item[1]['type'] == 'simple':
                filters = item[1]['columns']
            else:
                filters = np.concat([json_groups[c]['columns'] for c in item[1]['groups']]).tolist()

            try:
                mask_privileged, mask_underprivileged = get_mask(input_data, filters)
                input_data_underprivileged = input_data[mask_underprivileged].copy()
                input_data_privileged = input_data[mask_privileged].copy()

                mask_underprivileged = input_data_underprivileged[label_column].isin(positive_output)
                mask_privileged = input_data_privileged[label_column].isin(positive_output)
                # item is to convert np float64 to native python float.
                data_underprivileged = mask_underprivileged.sum() / mask_underprivileged.shape[0]
                p_positive = np.round(data_underprivileged*100, 1).item()
                data_privileged = mask_privileged.sum() / mask_privileged.shape[0]
                DI = np.round(data_underprivileged / data_privileged, 2).item()
                normalized_risk = self.normalize_value(DI)
                bias_level = self.get_bias_level(normalized_risk)
                result_list.update({group: {'DI': DI,
                                            '%_positive': p_positive,
                                            'normalized_risk': normalized_risk,
                                            'bias_level': bias_level}})
            except KeyError:
                # Captura cualquier ValueError lanzado por get_benchmarking
                logger.error(f"Group no present in data: '{group}'")
                result_list.update({group: {'DI': 0,
                                            '%_positive': 0,
                                            'normalized_risk': None,
                                            'bias_level': self.get_bias_level(0),
                                            'error': 'group no present in data.'}})
        logger.info(f"Completed: '{self.__str__()}'")
        return result_list

    def normalize_value(self, value):
        x = value
        normalized = 0
        if 0.8 < x < 1.2:
            if x == 1:
                normalized = 100
            elif x < 1:
                normalized = (x - 0.8) * (100 - 80) / (1 - 0.8) + 80
            else:
                normalized = (x - 1) * (80 - 100) / (1.2 - 1) + 100
        elif 1.2 <= x < 2:
            normalized = (x - 1.2) * (60 - 80) / (2 - 1.2) + 80
        elif x >= 2:
            normalized = 0
        elif 0.5 < x <= 0.8:
            normalized = (x - 0.5) * (80 - 60) / (0.8 - 0.5) + 60
        elif x <= 0.5:
            normalized = x * (60 - 0) / 0.5
        return normalized



# run test python -m unittest discover tests

import logging
import unittest
from eticas.data.loaders import load_dataset
from unittest.mock import patch
logger = logging.getLogger(__name__)


logging.basicConfig(
    level=logging.ERROR,
    format='[%(levelname)s] %(name)s - %(message)s'
)


class TestDataLoader(unittest.TestCase):

    def test_df_load_csv(self):
        expected = load_dataset('tests/loader_files/test_load.csv')
        self.assertTrue(expected.shape[0] > 0)

    def test_df_load_pickle(self):
        expected = load_dataset('tests/loader_files/test_load.pkl')
        self.assertTrue(expected.shape[0] > 0)

    def test_df_load_pickle_2(self):
        expected = load_dataset('tests/loader_files/test_load.pickle')
        self.assertTrue(expected.shape[0] > 0)

    def test_df_load_parquet(self):
        expected = load_dataset('tests/loader_files/test_load.parquet')
        self.assertTrue(expected.shape[0] > 0)

    def test_file_not_found(self):
        file_path = "non_existent_file.csv"
        with self.assertRaises(FileNotFoundError) as context:
            load_dataset(file_path)
        self.assertIn(f"The file '{file_path}' does not exist.", str(context.exception))

    def test_unsupported_extension(self):
        file_path = "file.unsupported"
        with patch("os.path.exists", return_value=True):
            with self.assertRaises(ValueError) as context:
                load_dataset(file_path)

        self.assertIn("Extension '.unsupported' is not supported.", str(context.exception))


if __name__ == '__main__':
    unittest.main()



"""
labeled_audit.py
=================

Implements an audit for the labeled phase of a model.
"""

from .base_audit import BaseAudit
from ..data.loaders import load_dataset

from ..metrics.da_inconsistency import Da_inconsistency
from ..metrics.da_positive import Da_positive
from ..metrics.dxa_inconsistency import Dxa_inconsistency
from ..metrics.da_informative import Da_informative
from ..metrics.d_statisticalparity import D_statisticalparity
from ..metrics.d_parity import D_parity
from ..metrics.performance import Performance
from ..metrics.d_equalodds import D_equalodds
from ..metrics.d_calibrated import D_calibrated
from ..metrics.da_fairness import Da_fairness
from ..metrics.disparate_impact import DisparateImpact
import logging
logger = logging.getLogger(__name__)


class LabeledAudit(BaseAudit):
    """
    Audit for the labeled / labeled phase of a model.
    """

    def run_audit(self,
                  dataset_path: str,
                  label_column: str,
                  output_column: str = None,
                  positive_output: list = None):
        """
        Performs checks related to the labeled / labeled  dataset and process.

       :param dataset_path: path to labeled dataset.
       :param label_column: Name of the column containing the target.
       :param output_column: Name of the column containing the prediction / classification.
       :param positive_output: Values of the column_output consider as positive.

        Returns
        -------
       :return: dict. The result of the labeled audit.
        """
        # Example: Check for data imbalance, missing values, etc.
        logger.info(f"Running labeled audit on model '{self.model.model_name}'")
        input_data = load_dataset(dataset_path)
        input_data = input_data.dropna()
        logger.info(f"labeled data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("labeled dataset shape is 0.")

        labeled_log = {}
        labeled_log.update({'da_inconsistency': Da_inconsistency().compute(input_data,
                                                                           self.model.sensitive_attributes)})
        labeled_log.update({'da_positive': Da_positive().compute(input_data,
                                                                 self.model.sensitive_attributes,
                                                                 label_column,
                                                                 positive_output)})
        labeled_log.update({'dxa_inconsistency': Dxa_inconsistency().compute(input_data,
                                                                             self.model.sensitive_attributes,
                                                                             self.model.features)})
        labeled_log.update({'da_informative': Da_informative().compute(input_data,
                                                                       self.model.sensitive_attributes,
                                                                       label_column,
                                                                       self.model.features)})
        labeled_log.update({'da_fairness': Da_fairness().compute(input_data,
                                                                 self.model.sensitive_attributes,
                                                                 self.model.features,
                                                                 label_column,
                                                                 positive_output)})
        labeled_log.update({'d_statisticalparity': D_statisticalparity().compute(input_data,
                                                                                 self.model.sensitive_attributes,
                                                                                 label_column,
                                                                                 positive_output)})
        labeled_log.update({'d_parity': D_parity().compute(input_data,
                                                           self.model.sensitive_attributes,
                                                           label_column,
                                                           positive_output)})
        labeled_log.update({'poor_performance': Performance().compute(input_data,
                                                                      self.model.sensitive_attributes,
                                                                      label_column,
                                                                      positive_output,
                                                                      output_column)})
        labeled_log.update({'d_equalodds': D_equalodds().compute(input_data,
                                                                 self.model.sensitive_attributes,
                                                                 label_column,
                                                                 positive_output, output_column)})
        labeled_log.update({'d_calibrated': D_calibrated().compute(input_data,
                                                                   self.model.sensitive_attributes,
                                                                   label_column,
                                                                   positive_output, output_column)})
        labeled_log.update({'impact_ratio': DisparateImpact().compute(input_data,
                                                                      self.model.sensitive_attributes,
                                                                      label_column,
                                                                      positive_output)})
        return labeled_log



"""
production_audit.py
===================

Implements an audit for the production (or testing) phase of a model.
"""


from .base_audit import BaseAudit
from ..data.loaders import load_dataset

from ..metrics.da_inconsistency import Da_inconsistency
from ..metrics.da_positive import Da_positive
from ..metrics.dxa_inconsistency import Dxa_inconsistency
from ..metrics.da_informative import Da_informative
from ..metrics.d_statisticalparity import D_statisticalparity
from ..metrics.d_parity import D_parity
from ..metrics.da_fairness import Da_fairness
from ..metrics.disparate_impact import DisparateImpact
import logging
logger = logging.getLogger(__name__)


class UnlabeledAudit(BaseAudit):
    """
    Audit for the production/test phase of a model.
    """
    def run_audit(self,
                  dataset_path: str,
                  output_column: str = None,
                  positive_output: list = None):
        """
        Performs checks related to the unlabeled / production  dataset and process.

       :param dataset_path: path to production dataset.
       :param output_column: Name of the column containing the prediction / classification.
       :param positive_output: Values of the column_output consider as positive.

        Returns
        -------
       :return: dict. The result of the unlabeled audit.
        """
        # Example: Check for data imbalance, missing values, etc.
        logger.info(f"Running production audit on model '{self.model.model_name}'")
        input_data = load_dataset(dataset_path)
        input_data = input_data.dropna()
        logger.info(f"production data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("production dataset shape is 0.")

        unlabeled_log = {}
        unlabeled_log.update({'da_inconsistency': Da_inconsistency().compute(input_data,
                                                                             self.model.sensitive_attributes)})
        unlabeled_log.update({'da_positive': Da_positive().compute(input_data,
                                                                   self.model.sensitive_attributes,
                                                                   output_column,
                                                                   positive_output)})
        unlabeled_log.update({'dxa_inconsistency': Dxa_inconsistency().compute(input_data,
                                                                               self.model.sensitive_attributes,
                                                                               self.model.features)})
        unlabeled_log.update({'da_informative': Da_informative().compute(input_data,
                                                                         self.model.sensitive_attributes,
                                                                         output_column,
                                                                         self.model.features)})
        unlabeled_log.update({'da_fairness': Da_fairness().compute(input_data,
                                                                   self.model.sensitive_attributes,
                                                                   self.model.features,
                                                                   output_column,
                                                                   positive_output)})
        unlabeled_log.update({'d_statisticalparity': D_statisticalparity().compute(input_data,
                                                                                   self.model.sensitive_attributes,
                                                                                   output_column,
                                                                                   positive_output)})
        unlabeled_log.update({'d_parity': D_parity().compute(input_data,
                                                             self.model.sensitive_attributes,
                                                             output_column,
                                                             positive_output)})
        unlabeled_log.update({'impact_ratio': DisparateImpact().compute(input_data,
                                                                        self.model.sensitive_attributes,
                                                                        output_column,
                                                                        positive_output)})
        return unlabeled_log



import numpy as np
from .base_metric import BaseFairnessMetric

from sklearn.model_selection import train_test_split
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import accuracy_score

import pandas as pd
from ..utils.helpers import get_mask
import logging
logger = logging.getLogger(__name__)


class Tdx_inconsistency(BaseFairnessMetric):
    """
    Metric Drift
    """

    def compute(self, sensitive_attrs: dict = None,
                input_features: list = None,
                input_data_dev: str = None,
                output_column_dev: str = None,
                positive_output_dev: list = [1],
                input_data_prod: str = None,
                output_column_prod: str = None,
                positive_output_prod: list = [1]):
        """
        Parameters
        ----------
        :param sensitive_attrs: The sensitive attributes
                                (e.g., 'gender', 'race') relevant to fairness.
        :param input_features: List of features use as input to get the output.
        :param input_data: Dataset which should contain the sensitive_attrs
                           and input_features.
        :param output_column: Name of the column containing the target.
        :param positive_output: Values of the column_output
                                consider as positive.
        Returns
        ----------
        :return: A dictionary containing the results for each group
        """
        self.validate_parameters(
            sensitive_attrs=sensitive_attrs,
            input_features=input_features,
            input_data_dev=input_data_dev,
            output_column_dev=output_column_dev,
            positive_output_dev=positive_output_dev,
            input_data_prod=input_data_prod,
            output_column_prod=output_column_prod,
            positive_output_prod=positive_output_prod
        )
        column_output = 'inconsistency'
        # Set random seed
        np.random.seed(123)
        input_data_dev[column_output] = 0
        input_data_prod[column_output] = 1
        len_dev = input_data_dev.shape[0]
        len_prod = input_data_prod.shape[0]
        total_len = np.sum([len_dev, len_prod])
        proportion_data = np.max([len_dev, len_prod]) / total_len

        json_groups = sensitive_attrs

        cols_dev = input_data_dev.columns
        cols_prod = input_data_prod.columns
        sensitive_columns = self.extract_sensitive_columns(json_groups)

        data_columns = set(cols_dev) & set(cols_prod)
        sensitive_columns = set(sensitive_columns) & set(data_columns)
        sensitive_columns = list(sensitive_columns)

        train_columns_0 = set(cols_dev) & set(input_features)
        train_columns_1 = set(cols_prod) & set(input_features)

        train_columns = set(train_columns_0) & set(train_columns_1)
        train_columns = list(train_columns)
        result_list = {}

        feat_columns = train_columns + sensitive_columns
        if len(feat_columns) == 0:
            logger.error("length 0 of train and sensitive features is 0.")
            return {'error': 'length 0 of train and sensitive features is 0.'}
        try:
            group = 'overall'
            data_dev = input_data_dev[feat_columns + [column_output]]
            data_prod = input_data_prod[feat_columns + [column_output]]
            data = pd.concat([data_dev, data_prod])
            # Split data (70% train, 30% test) based on the sensitive attribute
            train_data, test_data = train_test_split(data, test_size=0.3,
                                                     stratify=data[column_output],
                                                     random_state=123)
            logistic_base = LogisticRegression(random_state=123)
            logistic_base.fit(train_data[feat_columns], train_data[column_output])
            predictions_base = logistic_base.predict(test_data[feat_columns])
            accuracy_base = np.round(accuracy_score(test_data[column_output],
                                                    predictions_base), 4).item()
            rate = np.round(accuracy_base / proportion_data, 4).item()
            normalized_risk = np.round(self.normalize_value(rate), 4).item()
            result_list.update({group: {
                                'accuracy': np.round(accuracy_base, 4).item(),
                                'proportion': np.round(proportion_data, 4).item(),
                                'rate': rate,
                                'normalized_risk': normalized_risk,
                                'bias_level': self.get_bias_level(normalized_risk)
                                }})
        except ValueError:
            # Captura cualquier ValueError lanzado por get_benchmarking
            logger.error(f"Group no present in data: '{group}'")
            result_list.update({group: {
                                        'accuracy': 0,
                                        'proportion': 0,
                                        'rate': 0,
                                        'normalized_risk': 100,
                                        'bias_level': self.get_bias_level(100),
                                        'error': 'only one value target.'
                                        }})
        for item in json_groups.items():
            data_dev = input_data_dev
            data_prod = input_data_prod
            group = item[0]
            sensitive_columns = []

            try:
                if item[1]['type'] == 'simple':

                    filters = item[1]['columns']
                    mask_privileged, mask_underprivileged = get_mask(input_data_dev, filters)
                    data_dev[item[1]['columns'][0]['name']] = mask_underprivileged.astype(int)
                    mask_privileged_2, mask_underprivileged_2 = get_mask(input_data_prod, filters)
                    data_prod[item[1]['columns'][0]['name']] = mask_underprivileged.astype(int)
                    sensitive_columns.append(item[1]['columns'][0]['name'])

                else:
                    filters = np.concat([json_groups[c]['columns'] for c in item[1]['groups']]).tolist()
                    for filter in filters:
                        mask_privileged, mask_underprivileged = get_mask(input_data_dev, [filter])
                        data_dev[filter['name']] = mask_underprivileged.astype(int)
                        mask_privileged_2, mask_underprivileged_2 = get_mask(input_data_prod, [filter])
                        data_prod[filter['name']] = mask_underprivileged.astype(int)
                        sensitive_columns.append(filter['name'])
                feat_columns = train_columns + sensitive_columns

                data_dev = data_dev[mask_underprivileged][feat_columns + [column_output]]
                data_prod = data_prod[mask_underprivileged_2][feat_columns + [column_output]]
                data = pd.concat([data_dev, data_prod])
                # Split data (70% train, 30% test) based on the sensitive attribute
                train_data, test_data = train_test_split(data, test_size=0.3,
                                                         stratify=data[column_output],
                                                         random_state=123)
                logistic_base = LogisticRegression(random_state=123)
                logistic_base.fit(train_data[feat_columns], train_data[column_output])
                predictions_base = logistic_base.predict(test_data[feat_columns])
                accuracy_base = np.round(accuracy_score(test_data[column_output], predictions_base), 4).item()
                rate = np.round(accuracy_base / proportion_data, 4).item()
                normalized_risk = np.round(self.normalize_value(rate), 4).item()
                result_list.update({group: {
                                            'accuracy': np.round(accuracy_base, 4).item(),
                                            'proportion': np.round(proportion_data, 4).item(),
                                            'rate': rate,
                                            'normalized_risk': normalized_risk,
                                            'bias_level': self.get_bias_level(normalized_risk)
                                             }})
            except KeyError:
                # Captura cualquier ValueError lanzado por get_benchmarking
                logger.error(f"Group no present in data: '{group}'")
                result_list.update({group: {
                                            'accuracy': 0,
                                            'proportion': 0,
                                            'rate': 0,
                                            'normalized_risk': None,
                                            'bias_level': self.get_bias_level(0),
                                            'error': 'group no present in data.'
                                            }})
            except ValueError:
                # Captura cualquier ValueError lanzado por get_benchmarking
                logger.error(f"Only one value target: '{group}'")
                result_list.update({group: {
                                            'accuracy': 0,
                                            'proportion': 0,
                                            'rate': 0,
                                            'normalized_risk': 100,
                                            'bias_level': self.get_bias_level(100),
                                            'error': 'only one value target.'
                                            }})
        logger.info(f"Completed: '{self.__str__()}'")
        return result_list

    def extract_sensitive_columns(self, json_groups):

        columns = []

        for key, value in json_groups.items():
            column_data = value.get('columns', [])
            # Si los elementos son diccionarios, extraer el 'name'
            if isinstance(column_data, list) and len(column_data) > 0 and isinstance(column_data[0], dict):
                columns.extend([col['name'] for col in column_data])
            # Si son solo strings en una lista, los añadimos directamente
            elif isinstance(column_data, list) and all(isinstance(col, str) for col in column_data):
                columns.extend(column_data)

        return columns

    def normalize_value(self, value):
        accuracy = value - 1
        normalized = 0
        if accuracy <= 0:
            normalized = 100
        elif accuracy < 0.1:
            normalized = 100 - (abs(accuracy) / 0.1) * 20
        elif 0.1 <= accuracy < 0.2:
            normalized = 80 - ((accuracy - 0.1) / 0.1) * 20
        elif accuracy >= 0.2:
            normalized = 60 - ((accuracy - 0.2) / 0.8) * 60

        return normalized



import numpy as np
from .base_metric import BaseFairnessMetric
from ..utils.helpers import get_mask
import logging
logger = logging.getLogger(__name__)


class Dxa_inconsistency(BaseFairnessMetric):
    """
    Metric D(X | A)-inconsistency
    It is possible to predict a sensitive attribute based on the non-sensitive attributes in the data.
    This means that the sensitive attribute (i.e. gender) can be inferred from the other attributes,
    which means that there are proxies in the model.
    """

    def compute(self,
                input_data,
                sensitive_attrs=None,
                input_features=None):
        """
        Parameters
        ----------
       :param input_data: Dataset which should contain the columns in param sensitive_attrs.
       :param sensitive_attrs: The sensitive attributes (e.g., 'gender', 'race') relevant to fairness.
       :param input_features: List of features use as input to get the output.
        Returns
        ----------
       :return: A dictionary containing the results for each group
        """
        self.validate_parameters(
            input_data=input_data,
            sensitive_attrs=sensitive_attrs,
            input_features=input_features,
        )
        np.random.seed(123)
        input_data = input_data.dropna()
        json_groups = sensitive_attrs

        # Split data (70% train, 30% test) based on the sensitive attribute
        train_columns = [c for c in input_data.columns if c in input_features]

        if len(train_columns) == 0:
            logger.error("Input features are not in dataset.")
            return {'error': "Input features are not in dataset."}
        result_list = {}

        for item in json_groups.items():
            group = item[0]

            if item[1]['type'] == 'simple':
                filters = item[1]['columns']
            else:
                filters = np.concat([json_groups[c]['columns'] for c in item[1]['groups']]).tolist()

            try:

                mask_privileged, mask_underprivileged = get_mask(input_data, filters)
                # columns features extract automaticaly quiting sensitive.
                # actual manul.
                data = input_data.copy()[train_columns]
                data['y'] = mask_underprivileged.astype(int)
                proportion_data = np.max([data['y'].sum(), data.shape[0] - data['y'].sum()]) / data['y'].shape[0]

                correlation_matrix = abs(data.corr())

                correlation = correlation_matrix.loc['y'][train_columns].max()
                rate = np.round(correlation, 4).item()
                normalized_risk = np.round(self.normalize_value(rate), 4).item()
                result_list.update({group: {'correlation':  np.round(correlation, 4).item(),
                                            'proportion': np.round(proportion_data, 4).item(),
                                            'rate': rate,
                                            'normalized_risk': normalized_risk,
                                            'bias_level': self.get_bias_level(normalized_risk)}})
            except KeyError:
                # Captura cualquier ValueError lanzado por get_benchmarking
                logger.error(f"Group no present in data: '{group}'")

                result_list.update({group: {'correlation': 0,
                                            'proportion': 0,
                                            'rate': 0,
                                            'normalized_risk': None,
                                            'bias_level': self.get_bias_level(0),
                                            'error': 'group no present in data.'}})
        logger.info(f"Completed: '{self.__str__()}'")
        return result_list

    def normalize_value(self, value):
        '''
        Normalize the value between 80 - 100 if correlation is lower than 0.4
        Under 80 if value is higher than 0.4
        '''
        if value <= 0.25:
            # Normalización entre 80 y 100
            return np.min([100 - (value / 0.25 * 20), 100]).item()
        else:
            # Normalización entre 80 y 0
            return np.max([80 - ((value - 0.25) / 0.75 * 80), 0]).item()



import numpy as np
from .base_metric import BaseFairnessMetric
from ..utils.helpers import get_mask, binarize_predictions
import logging
logger = logging.getLogger(__name__)


class D_calibrated(BaseFairnessMetric):
    """
    Metric D(A)-positive
    Calculates the distribution of a given group when label is positive.
    """

    def compute(self,
                input_data,
                sensitive_attrs=None,
                label_column: str = None,
                positive_output: list = [1],
                output_column: str = None):
        """
        Parameters
        ----------
        :param input_data: Dataset which should contain the columns in param sensitive_attrs.
        :param sensitive_attrs: The sensitive attributes (e.g., 'gender', 'race') relevant to fairness.
        :param label_column: Name of the column containing the target.
        :param output_column: Name of the column containing the prediction / classification.
        :param positive_output: Values of the column_output consider as positive.
        Returns
        ----------
        :return: A dictionary containing the results for each group
        """
        self.validate_parameters(
            input_data=input_data,
            sensitive_attrs=sensitive_attrs,
            label_column=label_column,
            positive_output=positive_output,
            output_column=output_column
        )
        if not input_data[label_column].isin([0, 1]).all():
            # Convert to binary
            input_data[label_column] = binarize_predictions(input_data[label_column].values)
        if not input_data[output_column].isin([0, 1]).all():
            # Convert to binary
            input_data[output_column] = binarize_predictions(input_data[output_column].values)
        json_groups = sensitive_attrs
        result_list = {}

        for item in json_groups.items():
            group = item[0]

            try:
                if item[1]['type'] == 'simple':
                    filters = item[1]['columns']
                else:
                    filters = np.concat([json_groups[c]['columns'] for c in item[1]['groups']]).tolist()

                mask_privileged, mask_underprivileged = get_mask(input_data, filters)
                data_underprivileged = input_data[mask_underprivileged].copy()
                mask_positive_under = data_underprivileged[label_column].isin(positive_output)
                data_underprivileged_true = data_underprivileged[mask_positive_under]
                data_underprivileged_false = data_underprivileged[~mask_positive_under]
                data_privileged = input_data[mask_privileged].copy()
                mask_positive_privileged = data_privileged[label_column].isin(positive_output)
                data_privileged_true = data_privileged[mask_positive_privileged]
                data_privileged_false = data_privileged[~mask_positive_privileged]
                mask_underprivileged_true = data_underprivileged_true[output_column].isin(positive_output)
                mask_underprivileged_false = data_underprivileged_false[output_column].isin(positive_output)
                mask_privileged_true = data_privileged_true[output_column].isin(positive_output)
                mask_privileged_false = data_privileged_false[output_column].isin(positive_output)
                underprivileged_true = mask_underprivileged_true.sum() / mask_underprivileged_true.shape[0]
                underprivileged_false = mask_underprivileged_false.sum() / mask_underprivileged_false.shape[0]
                privileged_true = mask_privileged_true.sum() / mask_privileged_true.shape[0]
                privileged_false = mask_privileged_false.sum() / mask_privileged_false.shape[0]
                ratio_true = np.round(underprivileged_true / privileged_true, 4).item()
                ratio_false = np.round(underprivileged_false / privileged_false, 4).item()

                result_list.update({group: {
                                        'true_calibrated': {
                                            'underprivileged': np.round(underprivileged_true, 4).item(),
                                            'privileged': np.round(privileged_true, 4).item(),
                                            'ratio_true': ratio_true,
                                            'normalized_risk': np.round(self.normalize_value(ratio_true), 4).item(),
                                            'bias_level': self.get_bias_level(self.normalize_value(ratio_true))
                                                                },
                                        'false_calibrated': {
                                            'underprivileged': np.round(underprivileged_false, 4).item(),
                                            'privileged': np.round(privileged_false, 4).item(),
                                            'ratio_false': ratio_false,
                                            'normalized_risk': np.round(self.normalize_value(ratio_false), 4).item(),
                                            'bias_level': self.get_bias_level(self.normalize_value(ratio_false))
                                                                }
                                                                }})
            except KeyError:
                # Captura cualquier ValueError lanzado por get_benchmarking
                logger.error(f"Group no present in data: '{group}'")
                result_list.update({group: {
                                            'error': 'group no present in data.',
                                            'true_calibrated': {
                                                'underprivileged': 0,
                                                'privileged': 0,
                                                'ratio_true': 0,
                                                'normalized_risk': None,
                                                'bias_level': self.get_bias_level(0)
                                                                    },
                                            'false_calibrated': {
                                                'underprivileged': 0,
                                                'privileged': 0,
                                                'ratio_false': 0,
                                                'normalized_risk': None,
                                                'bias_level': self.get_bias_level(0)
                                                                    }
                                                                    }})
        logger.info(f"Completed: '{self.__str__()}'")
        return result_list

    def normalize_value(self, value):
        x = value
        normalized = 0
        if 0.8 < x < 1.2:
            if x == 1:
                normalized = 100
            elif x < 1:
                normalized = (x - 0.8) * (100 - 80) / (1 - 0.8) + 80
            else:
                normalized = (x - 1) * (80 - 100) / (1.2 - 1) + 100
        elif 1.2 <= x < 2:
            normalized = (x - 1.2) * (60 - 80) / (2 - 1.2) + 80
        elif x >= 2:
            normalized = 0
        elif 0.5 < x <= 0.8:
            normalized = (x - 0.5) * (80 - 60) / (0.8 - 0.5) + 60
        elif x <= 0.5:
            normalized = x * (60 - 0) / 0.5
        return normalized



import numpy as np
from .base_metric import BaseFairnessMetric
import math
import logging
import pandas as pd
from ..utils.helpers import get_mask, binarize_predictions
logger = logging.getLogger(__name__)


class Da_fairness(BaseFairnessMetric):
    """
    Metric D(A)-fairness
    Calculate two metrics : Equality and Equity.
        Equality: Means that all groups are treated in same manner.
        Equity: Means that some group needs additional resources to get success.
    """

    def compute(self,
                input_data,
                sensitive_attrs=None,
                input_features=None,
                label_column=None,
                positive_output: list = [1]):
        """
        Parameters
        ----------
       :param input_data: Dataset which should contain the columns in param sensitive_attrs.
       :param sensitive_attrs: The sensitive attributes (e.g., 'gender', 'race') relevant to fairness.
       :param label_column: Name of the column containing the target.
       :param input_features: List of features use as input to get the output.
        Returns
        ----------
       :return: A dictionary containing the results for each group
        """
        self.validate_parameters(
            input_data=input_data,
            sensitive_attrs=sensitive_attrs,
            input_features=input_features,
            label_column=label_column,
            positive_output=positive_output,
        )

        input_data = input_data.dropna()

        if not input_data[label_column].isin([0, 1]).all():
            # Convert to binary
            input_data[label_column] = binarize_predictions(input_data[label_column].values)

        train_columns = [c for c in input_data.columns if c in input_features]

        if len(train_columns) == 0:
            logger.error("Input features are not in dataset.")
            return {'error': "Input features are not in dataset."}

        json_groups = sensitive_attrs

        n_ranges = 5
        for f in input_features:
            if input_data[f].unique().shape[0] > 10:
                input_data[f] = pd.qcut(input_data[f], q=n_ranges, labels=range(5))

        result_list = {}
        for item in json_groups.items():
            data = input_data.copy()
            group = item[0]
            s_columns = []

            try:

                if item[1]['type'] == 'simple':

                    filters = item[1]['columns']
                    mask_privileged, mask_underprivileged = get_mask(input_data, filters)
                    data[item[1]['columns'][0]['name']] = mask_underprivileged.astype(int)
                    s_columns.append(item[1]['columns'][0]['name'])

                else:
                    filters = np.concat([json_groups[c]['columns'] for c in item[1]['groups']]).tolist()
                    mask_privileged, mask_underprivileged = get_mask(input_data, filters)
                    data['complex'] = mask_underprivileged.astype(int)
                    s_columns.append('complex')

                column_filter = s_columns

                df_aux = data.groupby(
                    by=column_filter + input_features, observed=False).agg({label_column: ['count', 'sum']})
                df_aux_ideal = data.groupby(
                    by=input_features, observed=False).agg({label_column: ['count', 'sum']})
                df_aux.columns = df_aux.columns.droplevel(0)
                df_aux = df_aux.reset_index()
                combinations_s = df_aux[column_filter].value_counts().index.values
                df_aux = df_aux.set_index(column_filter + input_features)

                df_aux_ideal.columns = df_aux_ideal.columns.droplevel(0)
                df_aux_ideal = df_aux_ideal.reset_index()
                # combinations_f = df_aux_ideal[input_features].value_counts().index.values
                df_aux_ideal['px'] = df_aux_ideal['sum'] / df_aux_ideal['count']
                df_aux_ideal = df_aux_ideal.sort_values(by=['px']+input_features)
                df_aux_ideal = df_aux_ideal.set_index(input_features)
                df_aux_ideal['dx'] = [0] + (df_aux_ideal['count'].cumsum() / df_aux_ideal['count'].sum()).tolist()[:-1]

                df_aux['px'] = df_aux['sum'] / df_aux['count']

                n_group = combinations_s.shape[0]
                # groups = [str(column_filter) + str(s) for s in combinations_s]
                # combinations = [[s + f for s in combinations_s] for f in combinations_f]
                df_aux = df_aux.reset_index().set_index(input_features)

                for cs, n in zip(combinations_s, range(n_group)):
                    condition = True
                    for feature, value in zip(column_filter, cs):
                        condition &= (df_aux[feature] == value)
                    filtered_aux = df_aux[condition]
                    df_aux_ideal = df_aux_ideal.merge(
                        filtered_aux, how='outer', left_index=True,
                        right_index=True, suffixes=["", "_"+str(n)]).fillna(0)
                    df_aux_ideal = df_aux_ideal.reset_index().sort_values(by=['px']+input_features)
                    df_aux_ideal = df_aux_ideal.set_index(input_features)
                    df_aux_ideal['dx_' + str(n)] = (
                        [0] +
                        (
                            df_aux_ideal['count_' + str(n)].cumsum() /
                            df_aux_ideal['count_' + str(n)].sum()
                        ).tolist()[:-1]
                    )
                n_p = -1
                p_max = 0
                d_max = 0
                for n in range(n_group):
                    p_aux = df_aux_ideal['px_'+str(n)].max()
                    d_aux = df_aux_ideal['dx_'+str(n)].max()
                    if p_aux > p_max:
                        p_max = p_aux
                        d_max = d_aux
                        n_p = n
                    elif p_aux == p_max:
                        if d_aux < d_max:
                            p_max = p_aux
                            d_max = d_aux
                            n_p = n

                # df_f = pd.DataFrame(columns=['group', 'reference', 'EQI', 'EQA', 'F'])
                for n in range(n_group):
                    if n != n_p:
                        eqi = (df_aux_ideal['dx_'+str(n_p)] - df_aux_ideal['dx_'+str(n)]).values
                        eqa = (df_aux_ideal['px_'+str(n_p)] - df_aux_ideal['px_'+str(n)]).values

                        # EQI = np.round(eqi.mean(),2)
                        EQI = abs((eqi * (df_aux_ideal['count'] / df_aux_ideal['count'].sum()).values).sum())
                        EQI = np.round(EQI, 2)
                        # EQA = np.round(eqa.mean(),2)
                        EQA = abs((eqa * (df_aux_ideal['count'] / df_aux_ideal['count'].sum()).values).sum())
                        EQA = np.round(EQA, 2)
                        F = np.round(math.sqrt(EQA**2 + EQI**2), 2)
                        EQA_norm = np.round(100*(1-EQA), 2)
                        EQI_norm = np.round(100*(1-EQI), 2)
                        normalized_risk = np.round(100*(1-F), 2)
                        # df_f.loc[n] = [groups[n], groups[n_p], EQI, EQA, F]
                result_list.update({group: {
                                            'equality': EQA.item(),
                                            'equity': EQI.item(),
                                            'fairness': F.item(),
                                            'equality_norm': EQA_norm.item(),
                                            'equity_norm': EQI_norm.item(),
                                            'normalized_risk': normalized_risk.item(),
                                            'bias_level': self.get_bias_level(normalized_risk)}})
            except KeyError:
                # Captura cualquier ValueError lanzado por get_benchmarking
                logger.error(f"Group no present in data: '{group}'")
                result_list.update({group: {
                                            'equality': 0,
                                            'equity': 0,
                                            'fairness': 0,
                                            'equality_norm': 0,
                                            'equity_norm': 0,
                                            'normalized_risk': None,
                                            'bias_level': self.get_bias_level(0),
                                            'error': 'group no present in data.'}})
        logger.info(f"Completed: '{self.__str__()}'")
        return result_list



from .base_metric import BaseFairnessMetric
from ..utils.helpers import binarize_predictions
import logging
logger = logging.getLogger(__name__)


class DisparateImpact(BaseFairnessMetric):
    """
    Calculates the impact ratio.
    """

    def compute(self,
                input_data,
                sensitive_attrs=None,
                label_column: str = None,
                positive_output: list = [1]):
        """
        Parameters
        ----------
        :param input_data: Dataset which should contain the columns in param sensitive_attrs.
        :param sensitive_attrs: The sensitive attributes (e.g., 'gender', 'race') relevant to fairness.
        :param label_column: Name of the column containing the target.
        :param positive_output: Values of the column_output consider as positive.
        Returns
        ----------
        :return: A dictionary containing the results for each group
        """
        self.validate_parameters(
            input_data=input_data,
            sensitive_attrs=sensitive_attrs,
            label_column=label_column,
            positive_output=positive_output
        )

        input_data = input_data.dropna()

        if not input_data[label_column].isin([0, 1]).all():
            # Convert to binary
            input_data[label_column] = binarize_predictions(input_data[label_column].values)
        json_groups = sensitive_attrs
        result_list = {}

        for item in json_groups.items():
            group = item[0]

            column = []
            if item[1]['type'] == 'simple':
                column.append(item[1]['columns'][0]['name'])
            if item[1]['type'] == 'complex':
                for g in item[1]['groups']:
                    column.append(json_groups[g]['columns'][0]['name'])
            if all([c in input_data.columns for c in column]):
                df_aux = (
                    input_data.groupby(column)
                    .agg(
                        **{
                            '# of Applicants': (label_column, 'count'),  # Conteo
                            'Selection Rate': (label_column, 'mean')     # Promedio
                        }
                    )
                    .reset_index()
                )
                df_aux['Impact Ratio'] = df_aux['Selection Rate'] / df_aux['Selection Rate'].max()
                df_aux['Selection Rate'] = (df_aux['Selection Rate'] * 100).map(lambda x: f"{x:.1f}%")

                n_pass = df_aux[df_aux['Impact Ratio'] >= 0.8].shape[0]
                df_aux['Impact Ratio'] = (df_aux['Impact Ratio']).map(lambda x: f"{x:.3f}")

                # df_merge = pd.DataFrame(input_data[column].drop_duplicates().values.tolist(), columns=[column])
                # df_aux = df_aux.merge(df_merge, how='right', left_on=column, right_on=column).fillna(0)

                mask = df_aux['# of Applicants'] <= 0
                df_aux.loc[mask, 'Selection Rate'] = 0
                df_aux.loc[mask, 'Impact Ratio'] = 0

                aux_unknown = input_data.shape[0] - df_aux['# of Applicants'].sum()

                result_list.update({group: {'df_result': df_aux.to_json(orient='split'),
                                            'unknown': aux_unknown.item(),
                                            'pass': n_pass,
                                            'total_filter': df_aux.shape[0] - mask.sum().item(),
                                            'total': df_aux.shape[0]}})
            else:
                result_list.update({group: {'df_result': None,
                                            'unknown': None,
                                            'pass': None,
                                            'total_filter': None,
                                            'total': None}})
        logger.info(f"Completed: '{self.__str__()}'")
        return result_list



# run test python -m unittest discover tests

import logging
from eticas.metrics.d_calibrated import D_calibrated
from eticas.metrics.d_equalodds import D_equalodds
from eticas.metrics.d_parity import D_parity
from eticas.metrics.d_statisticalparity import D_statisticalparity
from eticas.metrics.da_fairness import Da_fairness
from eticas.metrics.da_inconsistency import Da_inconsistency
from eticas.metrics.da_informative import Da_informative
from eticas.metrics.da_positive import Da_positive
from eticas.metrics.dxa_inconsistency import Dxa_inconsistency
from eticas.metrics.performance import Performance
from eticas.metrics.tdx_inconsistency import Tdx_inconsistency
from eticas.data.loaders import load_dataset
from eticas.metrics.disparate_impact import DisparateImpact
import unittest

sensitive_attributes = {'sex': {'columns': [
    {
        "name": "sex",
        "underprivileged": [2]
    }
],
    'type': 'simple'},
    'ethnicity': {'columns': [
        {
            "name": "ethnicity",
            "privileged": [1]
        }
    ],
    'type': 'simple'},
    'age': {'columns': [
        {
            "name": "age",
            "privileged": [3, 4]
        }
    ],
    'type': 'simple'},
    'sex_ethnicity': {'groups': ["sex", "ethnicity"],
                      'type': 'complex'}}

sensitive_attributes_error = {'sex': {'columns': [
                                            {
                                                "name": "sex",
                                                "underprivileged": [2]
                                            }
                                        ], 'type': 'simple'},
                              'ethnicity': {'columns': [
                                                 {
                                                     "name": "ethnicity",
                                                     "privileged": [1]
                                                 }
                                        ], 'type': 'simple'},
                              'age': {'columns': [
                                            {
                                                 "name": "age",
                                                 "privileged": [3, 4]
                                            }
                                        ], 'type': 'simple'},
                              'error': {'columns': [
                                  {
                                      "name": "error",
                                      "privileged": [3, 4]
                                  }
                                  ],
                                  'type': 'simple'},
                              'sex_ethnicity': {'groups': [
                                                    "sex", "ethnicity"
                                                    ], 'type': 'complex'}}

features = ["feature_0", "feature_1", "feature_2"]

label_column = 'outcome'
output_column = 'predicted_outcome'
positive_output = [1]

logger = logging.getLogger(__name__)


class TestMetrics(unittest.TestCase):

    def test_da_inconsistency(self):

        input_data = load_dataset('files/example_training_binary_2.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")

        result = Da_inconsistency().compute(input_data, sensitive_attributes)

        expected_result = {
            'age': {
                'data': 44.8},
            'ethnicity': {
                'data': 40.0},
            'sex': {
                'data': 60.0},
            'sex_ethnicity': {
                'data': 25.0}
        }

        # --- Check keys ---
        self.assertIn("age", result)
        self.assertIn("ethnicity", result)
        self.assertIn("sex", result)
        self.assertIn("sex_ethnicity", result)

        # --- Check numeric values (using assertAlmostEqual for floats) ---
        self.assertAlmostEqual(
            result["age"]["data"],
            expected_result["age"]["data"],
            places=7,
            msg="da_inconsistency for age does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex"]["data"],
            expected_result["sex"]["data"],
            places=7,
            msg="da_inconsistency for sex does not match expected value"
        )
        self.assertAlmostEqual(
            result["ethnicity"]["data"],
            expected_result["ethnicity"]["data"],
            places=7,
            msg="da_inconsistency for ethnicity does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex_ethnicity"]["data"],
            expected_result["sex_ethnicity"]["data"],
            places=7,
            msg="da_positive for sex_ethnicity does not match expected value"
        )

    def test_da_positive(self):

        input_data = load_dataset('files/example_training_scoring.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")

        result = Da_positive().compute(input_data, sensitive_attributes,
                                       label_column, positive_output)
        input_data = load_dataset('files/example_training_binary_2.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")

        result = Da_positive().compute(input_data, sensitive_attributes,
                                       label_column, positive_output)

        expected_result = {
            'age': {
                'data': 45.2},
            'ethnicity': {
                'data': 39.8},
            'sex': {
                'data': 59.8},
            'sex_ethnicity': {
                'data': 24.7}
        }

        # --- Check keys ---
        self.assertIn("age", result)
        self.assertIn("ethnicity", result)
        self.assertIn("sex", result)
        self.assertIn("sex_ethnicity", result)

        # --- Check numeric values (using assertAlmostEqual for floats) ---
        self.assertAlmostEqual(
            result["age"]["data"],
            expected_result["age"]["data"],
            places=7,
            msg="da_positive for age does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex"]["data"],
            expected_result["sex"]["data"],
            places=7,
            msg="da_positive for sex does not match expected value"
        )
        self.assertAlmostEqual(
            result["ethnicity"]["data"],
            expected_result["ethnicity"]["data"],
            places=7,
            msg="da_positive for ethnicity does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex_ethnicity"]["data"],
            expected_result["sex_ethnicity"]["data"],
            places=7,
            msg="da_positive for sex_ethnicity does not match expected value"
        )

    def test_dxa_inconsistency(self):
        input_data = load_dataset('files/example_training_scoring.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")
        input_data = load_dataset('files/example_training_binary_2.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")

        result = Dxa_inconsistency().compute(input_data, sensitive_attributes, features)

        expected_result = {
            'age': {
                'normalized_risk': 99.28},
            'ethnicity': {
                'normalized_risk': 99.64},
            'sex': {
                'normalized_risk': 99.128},
            'sex_ethnicity': {
                'normalized_risk': 98.808}
        }

        # --- Check keys ---
        self.assertIn("age", result)
        self.assertIn("ethnicity", result)
        self.assertIn("sex", result)
        self.assertIn("sex_ethnicity", result)

        # --- Check numeric values (using assertAlmostEqual for floats) ---
        self.assertAlmostEqual(
            result["age"]["normalized_risk"],
            expected_result["age"]["normalized_risk"],
            places=7,
            msg="dxa_inconsistency for age does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex"]["normalized_risk"],
            expected_result["sex"]["normalized_risk"],
            places=7,
            msg="dxa_inconsistency for sex does not match expected value"
        )
        self.assertAlmostEqual(
            result["ethnicity"]["normalized_risk"],
            expected_result["ethnicity"]["normalized_risk"],
            places=7,
            msg="dxa_inconsistency for ethnicity does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex_ethnicity"]["normalized_risk"],
            expected_result["sex_ethnicity"]["normalized_risk"],
            places=7,
            msg="dxa_inconsistency for sex_ethnicity does not match expected value"
        )

    def test_da_informative(self):
        input_data = load_dataset('files/example_training_scoring.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")
        input_data = load_dataset('files/example_training_binary_2.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")

        result = Da_informative().compute(
            input_data, sensitive_attributes, label_column, features)

        expected_result = {
            'age': {
                'normalized_risk': 100.0},
            'ethnicity': {
                'normalized_risk': 100},
            'sex': {
                'normalized_risk': 99.3294},
            'sex_ethnicity': {
                'normalized_risk': 100.0}
        }

        # --- Check keys ---
        self.assertIn("age", result)
        self.assertIn("ethnicity", result)
        self.assertIn("sex", result)
        self.assertIn("sex_ethnicity", result)

        # --- Check numeric values (using assertAlmostEqual for floats) ---
        self.assertAlmostEqual(
            result["age"]["normalized_risk"],
            expected_result["age"]["normalized_risk"],
            places=7,
            msg="da_informative for age does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex"]["normalized_risk"],
            expected_result["sex"]["normalized_risk"],
            places=7,
            msg="da_informative for sex does not match expected value"
        )
        self.assertAlmostEqual(
            result["ethnicity"]["normalized_risk"],
            expected_result["ethnicity"]["normalized_risk"],
            places=7,
            msg="da_informative for ethnicity does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex_ethnicity"]["normalized_risk"],
            expected_result["sex_ethnicity"]["normalized_risk"],
            places=7,
            msg="da_informative for sex_ethnicity does not match expected value"
        )

        self.assertAlmostEqual(Da_informative().normalize_value(1.0), 100)
        self.assertAlmostEqual(Da_informative().normalize_value(1.09), 81.99999999999999)
        self.assertAlmostEqual(Da_informative().normalize_value(1.19), 62.000000000000014)
        self.assertAlmostEqual(Da_informative().normalize_value(1.21), 59.25)

    def test_d_statisticalparity(self):

        input_data = load_dataset('files/example_training_scoring.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")

        result = D_statisticalparity().compute(input_data, sensitive_attributes,
                                               label_column, positive_output)
        input_data = load_dataset('files/example_training_binary_2.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")

        result = D_statisticalparity().compute(input_data, sensitive_attributes,
                                               label_column, positive_output)

        expected_result = {
            'age': {
                'normalized_risk': 98.0},
            'ethnicity': {
                'normalized_risk': 100},
            'sex': {
                'normalized_risk': 100},
            'sex_ethnicity': {
                'normalized_risk': 98.0}
        }

        # --- Check keys ---
        self.assertIn("age", result)
        self.assertIn("ethnicity", result)
        self.assertIn("sex", result)
        self.assertIn("sex_ethnicity", result)

        # --- Check numeric values (using assertAlmostEqual for floats) ---
        self.assertAlmostEqual(
            result["age"]["normalized_risk"],
            expected_result["age"]["normalized_risk"],
            places=7,
            msg="d_statisticalparity for age does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex"]["normalized_risk"],
            expected_result["sex"]["normalized_risk"],
            places=7,
            msg="d_statisticalparity for sex does not match expected value"
        )
        self.assertAlmostEqual(
            result["ethnicity"]["normalized_risk"],
            expected_result["ethnicity"]["normalized_risk"],
            places=7,
            msg="d_statisticalparity for ethnicity does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex_ethnicity"]["normalized_risk"],
            expected_result["sex_ethnicity"]["normalized_risk"],
            places=7,
            msg="d_statisticalparity for sex_ethnicity does not match expected value"
        )

    def test_d_parity(self):
        input_data = load_dataset('files/example_training_scoring.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")

        result = D_parity().compute(input_data, sensitive_attributes,
                                    label_column, positive_output)
        input_data = load_dataset('files/example_training_binary_2.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")

        result = D_parity().compute(input_data, sensitive_attributes,
                                    label_column, positive_output)

        expected_result = {
            'age': {
                'normalized_risk': 98.0},
            'ethnicity': {
                'normalized_risk': 99.0},
            'sex': {
                'normalized_risk': 99.0},
            'sex_ethnicity': {
                'normalized_risk': 99.0}
        }

        # --- Check keys ---
        self.assertIn("age", result)
        self.assertIn("ethnicity", result)
        self.assertIn("sex", result)
        self.assertIn("sex_ethnicity", result)

        # --- Check numeric values (using assertAlmostEqual for floats) ---
        self.assertAlmostEqual(
            result["age"]["normalized_risk"],
            expected_result["age"]["normalized_risk"],
            places=7,
            msg="d_parity for age does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex"]["normalized_risk"],
            expected_result["sex"]["normalized_risk"],
            places=7,
            msg="d_parity for sex does not match expected value"
        )
        self.assertAlmostEqual(
            result["ethnicity"]["normalized_risk"],
            expected_result["ethnicity"]["normalized_risk"],
            places=7,
            msg="d_parity for ethnicity does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex_ethnicity"]["normalized_risk"],
            expected_result["sex_ethnicity"]["normalized_risk"],
            places=7,
            msg="d_parity for sex_ethnicity does not match expected value"
        )

        self.assertEqual(D_parity().normalize_value(0.8), 80)
        self.assertEqual(D_parity().normalize_value(1), 100)
        self.assertEqual(D_parity().normalize_value(1.2), 80)
        self.assertAlmostEqual(D_parity().normalize_value(0.9), 90)
        self.assertAlmostEqual(D_parity().normalize_value(1.1), 90)
        self.assertAlmostEqual(D_parity().normalize_value(1.5), 72.5)
        self.assertAlmostEqual(D_parity().normalize_value(1.8), 65)
        self.assertEqual(D_parity().normalize_value(2), 0)
        self.assertEqual(D_parity().normalize_value(3), 0)
        self.assertAlmostEqual(D_parity().normalize_value(0.65), 70)
        self.assertAlmostEqual(D_parity().normalize_value(0.70), 73.33333333333333)
        self.assertAlmostEqual(D_parity().normalize_value(0.5), 60)
        self.assertAlmostEqual(D_parity().normalize_value(0.3), 36)
        self.assertEqual(D_parity().normalize_value(0), 0)
        self.assertEqual(D_parity().normalize_value(-1), -120)

    def test_performance(self):
        input_data = load_dataset('files/example_training_scoring.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")

        result = Performance().compute(input_data, sensitive_attributes,
                                       label_column, positive_output, output_column)
        input_data = load_dataset('files/example_training_binary_2.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")

        result = Performance().compute(input_data, sensitive_attributes,
                                       label_column, positive_output, output_column)

        expected_result = {
            'age': {
                'normalized_risk': 57.588},
            'ethnicity': {
                'normalized_risk': 59.58},
            'sex': {
                'normalized_risk': 59.052},
            'sex_ethnicity': {
                'normalized_risk': 58.572}
        }

        # --- Check keys --
        self.assertIn("age", result)
        self.assertIn("ethnicity", result)
        self.assertIn("sex", result)
        self.assertIn("sex_ethnicity", result)

        # --- Check numeric values (using assertAlmostEqual for floats) ---
        self.assertAlmostEqual(
            result["age"]["normalized_risk"],
            expected_result["age"]["normalized_risk"],
            places=7,
            msg="performance for age does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex"]["normalized_risk"],
            expected_result["sex"]["normalized_risk"],
            places=7,
            msg="performance for sex does not match expected value"
        )
        self.assertAlmostEqual(
            result["ethnicity"]["normalized_risk"],
            expected_result["ethnicity"]["normalized_risk"],
            places=7,
            msg="performance for ethnicity does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex_ethnicity"]["normalized_risk"],
            expected_result["sex_ethnicity"]["normalized_risk"],
            places=7,
            msg="performance for sex_ethnicity does not match expected value"
        )

        self.assertEqual(Performance().normalize_value(0.9), 54.0)
        self.assertEqual(Performance().normalize_value(1.2), 73.33333333333333)
        self.assertEqual(Performance().normalize_value(1.5), 85.71428571428571)
        self.assertAlmostEqual(Performance().normalize_value(4), 100)

    def test_da_fairness(self):
        input_data = load_dataset('files/example_training_scoring.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")
        input_data = load_dataset('files/example_training_binary_2.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")

        result = Da_fairness().compute(input_data, sensitive_attributes,
                                       features, label_column, positive_output)

        expected_result = {
            'age': {
                'equity': 0.0,
                'equality': 0.01},
            'ethnicity': {
                'equity': 0.0,
                'equality': 0.0},
            'sex': {
                'equity': 0.01,
                'equality': 0.0},
            'sex_ethnicity': {
                'equity': 0.0,
                'equality': 0.01}
        }

        # --- Check keys ---
        self.assertIn("age", result)
        self.assertIn("ethnicity", result)
        self.assertIn("sex", result)
        self.assertIn("sex_ethnicity", result)

        # --- Check numeric values (using assertAlmostEqual for floats) ---
        self.assertAlmostEqual(
            result["age"]["equity"],
            expected_result["age"]["equity"],
            places=7,
            msg="da_fairness equity for age does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex"]["equity"],
            expected_result["sex"]["equity"],
            places=7,
            msg="da_fairness equity for sex does not match expected value"
        )
        self.assertAlmostEqual(
            result["ethnicity"]["equity"],
            expected_result["ethnicity"]["equity"],
            places=7,
            msg="da_fairness equity for ethnicity does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex_ethnicity"]["equity"],
            expected_result["sex_ethnicity"]["equity"],
            places=7,
            msg="da_fairness equity for sex_ethnicity does not match expected value"
        )

        # --- Check numeric values (using assertAlmostEqual for floats) ---
        self.assertAlmostEqual(
            result["age"]["equality"],
            expected_result["age"]["equality"],
            places=7,
            msg="da_fairness equality for age does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex"]["equality"],
            expected_result["sex"]["equality"],
            places=7,
            msg="da_fairness equality for sex does not match expected value"
        )
        self.assertAlmostEqual(
            result["ethnicity"]["equality"],
            expected_result["ethnicity"]["equality"],
            places=7,
            msg="da_fairness equality for ethnicity does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex_ethnicity"]["equality"],
            expected_result["sex_ethnicity"]["equality"],
            places=7,
            msg="da_fairness equality for sex_ethnicity does not match expected value"
        )

    def test_calibrated(self):

        input_data = load_dataset('files/example_training_scoring.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")
        result = D_calibrated().compute(input_data, sensitive_attributes,
                                        label_column, positive_output, output_column)
        input_data = load_dataset('files/example_training_binary_2.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")
        result = D_calibrated().compute(input_data, sensitive_attributes,
                                        label_column, positive_output, output_column)
        expected_result = {
            'age': {
                'true_calibrated': 96.82,
                'false_calibrated': 99.76},
            'ethnicity': {
                'true_calibrated': 98.6,
                'false_calibrated': 97.86},
            'sex': {
                'true_calibrated': 96.36,
                'false_calibrated': 98.69},
            'sex_ethnicity': {
                'true_calibrated': 95.01,
                'false_calibrated': 99.18}
        }

        # --- Check keys --
        self.assertIn("age", result)
        self.assertIn("ethnicity", result)
        self.assertIn("sex", result)
        self.assertIn("sex_ethnicity", result)

        # --- Check numeric values (using assertAlmostEqual for floats) ---
        self.assertAlmostEqual(
            result["age"]["true_calibrated"]["normalized_risk"],
            expected_result["age"]["true_calibrated"],
            places=7,
            msg="d_calibrated true for age does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex"]["true_calibrated"]["normalized_risk"],
            expected_result["sex"]['true_calibrated'],
            places=7,
            msg="d_calibrated true for sex does not match expected value"
        )
        self.assertAlmostEqual(
            result["ethnicity"]["true_calibrated"]["normalized_risk"],
            expected_result["ethnicity"]['true_calibrated'],
            places=7,
            msg="d_calibrated true for ethnicity does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex_ethnicity"]["true_calibrated"]["normalized_risk"],
            expected_result["sex_ethnicity"]['true_calibrated'],
            places=7,
            msg="d_calibrated true for sex_ethnicity does not match expected value"
        )
        self.assertAlmostEqual(
            result["age"]["false_calibrated"]["normalized_risk"],
            expected_result["age"]['false_calibrated'],
            places=7,
            msg="d_calibrated false for age does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex"]["false_calibrated"]["normalized_risk"],
            expected_result["sex"]['false_calibrated'],
            places=7,
            msg="d_calibrated false for sex does not match expected value"
        )
        self.assertAlmostEqual(
            result["ethnicity"]["false_calibrated"]["normalized_risk"],
            expected_result["ethnicity"]['false_calibrated'],
            places=7,
            msg="d_calibrated false for ethnicity does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex_ethnicity"]["false_calibrated"]["normalized_risk"],
            expected_result["sex_ethnicity"]['false_calibrated'],
            places=7,
            msg="d_calibrated false for sex_ethnicity does not match expected value"
        )

        self.assertEqual(D_calibrated().normalize_value(0.8), 80)
        self.assertEqual(D_calibrated().normalize_value(1), 100)
        self.assertEqual(D_calibrated().normalize_value(1.2), 80)
        self.assertAlmostEqual(D_calibrated().normalize_value(0.9), 90)
        self.assertAlmostEqual(D_calibrated().normalize_value(1.1), 90)
        self.assertAlmostEqual(D_calibrated().normalize_value(1.5), 72.5)
        self.assertAlmostEqual(D_calibrated().normalize_value(1.8), 65)
        self.assertEqual(D_calibrated().normalize_value(2), 0)
        self.assertEqual(D_calibrated().normalize_value(3), 0)
        self.assertAlmostEqual(D_calibrated().normalize_value(0.65), 70)
        self.assertAlmostEqual(D_calibrated().normalize_value(0.70), 73.33333333333333)
        self.assertAlmostEqual(D_calibrated().normalize_value(0.5), 60)
        self.assertAlmostEqual(D_calibrated().normalize_value(0.3), 36)
        self.assertEqual(D_calibrated().normalize_value(0), 0)
        self.assertEqual(D_calibrated().normalize_value(-1), -120)

    def test_equalodd(self):

        input_data = load_dataset('files/example_training_scoring.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")

        result = D_equalodds().compute(input_data, sensitive_attributes,
                                       label_column, positive_output, output_column)
        input_data = load_dataset('files/example_training_binary_2.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")

        result = D_equalodds().compute(input_data, sensitive_attributes,
                                       label_column, positive_output, output_column)

        expected_result = {
            'age': {
                'true_positive_rate': 96.82,
                'false_positive_rate': 99.76},
            'ethnicity': {
                'true_positive_rate': 98.6,
                'false_positive_rate': 97.86},
            'sex': {
                'true_positive_rate': 96.36,
                'false_positive_rate': 98.69},
            'sex_ethnicity': {
                'true_positive_rate': 95.01,
                'false_positive_rate': 99.18}
        }

        # --- Check keys --
        self.assertIn("age", result)
        self.assertIn("ethnicity", result)
        self.assertIn("sex", result)
        self.assertIn("sex_ethnicity", result)

        # --- Check numeric values (using assertAlmostEqual for floats) ---
        self.assertAlmostEqual(
            result["age"]["true_positive_rate"]["normalized_risk"],
            expected_result["age"]["true_positive_rate"],
            places=7,
            msg="d_equalodd true for age does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex"]["true_positive_rate"]["normalized_risk"],
            expected_result["sex"]['true_positive_rate'],
            places=7,
            msg="d_equalodd true for sex does not match expected value"
        )
        self.assertAlmostEqual(
            result["ethnicity"]["true_positive_rate"]["normalized_risk"],
            expected_result["ethnicity"]['true_positive_rate'],
            places=7,
            msg="d_equalodd true for ethnicity does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex_ethnicity"]["true_positive_rate"]["normalized_risk"],
            expected_result["sex_ethnicity"]['true_positive_rate'],
            places=7,
            msg="d_equalodd true for sex_ethnicity does not match expected value"
        )
        self.assertAlmostEqual(
            result["age"]["false_positive_rate"]["normalized_risk"],
            expected_result["age"]['false_positive_rate'],
            places=7,
            msg="d_equalodd false for age does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex"]["false_positive_rate"]["normalized_risk"],
            expected_result["sex"]['false_positive_rate'],
            places=7,
            msg="d_equalodd false for sex does not match expected value"
        )
        self.assertAlmostEqual(
            result["ethnicity"]["false_positive_rate"]["normalized_risk"],
            expected_result["ethnicity"]['false_positive_rate'],
            places=7,
            msg="d_equalodd false for ethnicity does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex_ethnicity"]["false_positive_rate"]["normalized_risk"],
            expected_result["sex_ethnicity"]['false_positive_rate'],
            places=7,
            msg="d_equalodd false for sex_ethnicity does not match expected value"
        )

        self.assertEqual(D_equalodds().normalize_value(0.8), 80)
        self.assertEqual(D_equalodds().normalize_value(1), 100)
        self.assertEqual(D_equalodds().normalize_value(1.2), 80)
        self.assertAlmostEqual(D_equalodds().normalize_value(0.9), 90)
        self.assertAlmostEqual(D_equalodds().normalize_value(1.1), 90)
        self.assertAlmostEqual(D_equalodds().normalize_value(1.5), 72.5)
        self.assertAlmostEqual(D_equalodds().normalize_value(1.8), 65)
        self.assertEqual(D_equalodds().normalize_value(2), 0)
        self.assertEqual(D_equalodds().normalize_value(3), 0)
        self.assertAlmostEqual(D_equalodds().normalize_value(0.65), 70)
        self.assertAlmostEqual(D_equalodds().normalize_value(0.70), 73.33333333333333)
        self.assertAlmostEqual(D_equalodds().normalize_value(0.5), 60)
        self.assertAlmostEqual(D_equalodds().normalize_value(0.3), 36)
        self.assertEqual(D_equalodds().normalize_value(0), 0)
        self.assertEqual(D_equalodds().normalize_value(-1), -120)

    def test_tdx_inconsistency(self):

        input_data = load_dataset('files/example_training_scoring.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")

        input_data = load_dataset('files/example_training_binary_2.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")

        result = Tdx_inconsistency().compute(sensitive_attributes, features,
                                             input_data, output_column, output_column,
                                             input_data, output_column, output_column)
        expected_result = {
            'age': {
                'normalized_risk': 100.0},
            'ethnicity': {
                'normalized_risk': 100.0},
            'sex': {
                'normalized_risk': 100.0},
            'sex_ethnicity': {
                'normalized_risk': 100.0}
        }

        # --- Check keys ---
        self.assertIn("age", result)
        self.assertIn("ethnicity", result)
        self.assertIn("sex", result)
        self.assertIn("sex_ethnicity", result)

        # --- Check numeric values (using assertAlmostEqual for floats) ---
        self.assertAlmostEqual(
            result["age"]["normalized_risk"],
            expected_result["age"]["normalized_risk"],
            places=7,
            msg="Tdx_inconsistency for age does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex"]["normalized_risk"],
            expected_result["sex"]["normalized_risk"],
            places=7,
            msg="Tdx_inconsistency for sex does not match expected value"
        )
        self.assertAlmostEqual(
            result["ethnicity"]["normalized_risk"],
            expected_result["ethnicity"]["normalized_risk"],
            places=7,
            msg="Tdx_inconsistency for ethnicity does not match expected value"
        )
        self.assertAlmostEqual(
            result["sex_ethnicity"]["normalized_risk"],
            expected_result["sex_ethnicity"]["normalized_risk"],
            places=7,
            msg="Tdx_inconsistency for sex_ethnicity does not match expected value"
        )

        self.assertAlmostEqual(Tdx_inconsistency().normalize_value(1.0), 100)
        self.assertAlmostEqual(Tdx_inconsistency().normalize_value(1.09), 81.99999999999999)
        self.assertAlmostEqual(Tdx_inconsistency().normalize_value(1.19), 62.000000000000014)
        self.assertAlmostEqual(Tdx_inconsistency().normalize_value(1.21), 59.25)

    def test_disparate_impact(self):

        input_data = load_dataset('files/example_training_scoring.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")

        DisparateImpact().compute(input_data, sensitive_attributes,
                                  label_column, positive_output)
        input_data = load_dataset('files/example_training_binary_2.csv')
        input_data = input_data.dropna()
        logger.info(f"Training data loaded '{input_data.shape}'")
        if input_data.shape[0] == 0:
            raise ValueError("Training dataset shape is 0.")

        DisparateImpact().compute(input_data, sensitive_attributes,
                                  label_column, positive_output)


if __name__ == '__main__':
    unittest.main()



# `setup.py`
from setuptools import setup, find_packages
import os
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


def read_requirements():
    req_path = os.path.join(os.path.dirname(__file__), "requirements.txt")
    with open(req_path, encoding="utf-8") as f:
        return [line.strip() for line in f if line.strip() and not line.startswith("#")]


setup(
    name="eticas",
    version="0.1.0",
    author="Eticas AI",
    author_email="it@eticas.ai",
    description="A Python library for calculating fairness metrics in ML models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/eticasai/eticas-audit",
    install_requires=[
        "numpy==2.1.2",
        "pandas==2.2.3",
        "scikit-learn==1.5.2",
        "pyarrow==19.0.0",
        "scipy==1.15.1",
        "coverage==7.6.10",
        "requests==2.32.3"
    ],
    packages=find_packages(include=['eticas', 'eticas.*'])
)

